parser grammar DuraC;

options {
  output=AST;
  tokenVocab=Dura;
}

@header {
package de.lmu.ifi.pms.cep.dura.compiler.antlr;

import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraParseException;
}

@members {
	private int virtualIdentifier = 0;
	
	public void displayRecognitionError(String[] tokenNames, RecognitionException e) {
	
		String hdr = getErrorHeader(e);
		String msg = getErrorMessage(e, tokenNames);
		
		throw new DuraParseException(msg, getSourceName(), e.line);
	}
	
	private CommonToken token(String text, int type, Token start) {
		CommonToken t = new CommonToken(type, text);
		t.setLine(start.getLine());
		return t;
	}
}

/* DURA Core*/

program
  :  preamble (def+=eventDefinition | def+=stateDefinition | def+=actionDefinition | def+=reactiveRule)*
  ->  ^(PROGRAM preamble ^(DEFINITIONS $def*))
  ;

preamble
  : pckg description load* imprt* export* (constDefinition | typeDefinition)*
  ->  ^(PREAMBLE pckg description load* imprt* export* constDefinition* typeDefinition*)
  ;

pckg
  : ('MODULE' path)?
  ->  ^(MODULE path?)
  ;
  
description
  : ('DESCRIPTION' STRING)?
  ->  ^(DESCRIPTION STRING?)
  ;
  
imprt
  : (IMPORT path)
  -> ^(IMPORT path)
  ;

load
	:	(LOAD path)
	-> ^(LOAD path)
	;
	
export
	:	(EXPORT path)
	-> ^(EXPORT path)
	;

/** 
 * constraints: 
 *   all schemaDefinition labels are pairwise distinct
 *   type name ID is unique
 *   no cyclic definitions
 */
typeDefinition 
  : visibilityModifier? TYPEDEF ID 'IS' basicType typeSupplement 'END'
  ->  ^(TYPEDEF[$ID] ^(MODIFIERS visibilityModifier?) ^(TYPE basicType typeSupplement))
  | visibilityModifier? TYPEDEF ID 'IS' schemaDefinition typeSupplement 'END'
  ->  ^(TYPEDEF[$ID] ^(MODIFIERS visibilityModifier?) ^(TYPE schemaDefinition typeSupplement))
  | visibilityModifier? TYPEDEF ID 'IS' '{' schemaDefinition (',' schemaDefinition)* '}' typeSupplement 'END'  
  ->  ^(TYPEDEF[$ID] ^(MODIFIERS visibilityModifier?) ^(TYPE schemaDefinition+ typeSupplement))
  ;

/** constraint: const name ID is unique */
constDefinition
  : visibilityModifier? CONSTDEF ID 'IS' constTerm 'END'      ->  ^(CONSTDEF[$ID] ^(MODIFIERS visibilityModifier?) constTerm)
  ;


/**
 * constraints:
 *   event definition schemaDefinition needs to be unique
 *   schema of derived events need to comply with schmaDenfinition
 */
eventDefinition
  : modifiers EVENT schemaDefinition schemaSupplement ('WITH' eventSpecification*)? 'END'
  ->  ^(EVENT modifiers ^(SCHEMA schemaDefinition schemaSupplement) eventSpecification*)
  ;

/** constraint: stateful object definition schemaDefinition is unique */
stateDefinition
  : modifiers STATEFUL_OBJECT schemaDefinition schemaSupplement 'END'
  ->  ^(STATEFUL_OBJECT modifiers ^(SCHEMA schemaDefinition schemaSupplement))
  ;

/** constraint: action definition schemaDefinition is unique */
actionDefinition
  : modifiers ACTION schemaDefinition schemaSupplement 'END'
  ->  ^(ACTION modifiers ^(SCHEMA schemaDefinition schemaSupplement))
  ;

/** constraint: query is range restricted with respect to term */
eventSpecification
  : DETECT term 'ON' eventQuery 'END'
  ->  ^(DETECT ^(HEAD term) ^(BODY eventQuery))
  ;
  
eventSpecifications
	:	eventSpecification*
  ;

/** constraint: action is range restricted with respect to query */
reactiveRule
  : 'ON' eventQuery 'DO' action 'END'
  ->  ^(DO ^(BODY eventQuery) ^(HEAD action))
  ;
  

/** constaint: at least one event query in every disjunct */
eventQuery
  : CONJUNCTION '{' flatEventQuery (','? flatEventQuery)* ','? '}' querySupplement ->  ^(CONJUNCTION flatEventQuery+ querySupplement)
  | DISJUNCTION '{' flatEventQuery (','? flatEventQuery)* ','? '}' querySupplement  ->  ^(DISJUNCTION flatEventQuery+ querySupplement)
  | flatEventQuery                                                          ->  ^(CONJUNCTION flatEventQuery ^(SUPPLEMENT))
  ;

flatEventQuery
  : NEGATION '{' atomicEventQuery '}' negationSupplement  ->  ^(NEGATION atomicEventQuery negationSupplement)
  | NEGATION atomicEventQuery ->  ^(NEGATION atomicEventQuery SUPPLEMENT)
  | atomicEventQuery
  ;

atomicEventQuery
  : EVENT_IDENTIFIER ID ':' qualifiedTerm querySupplement ->  ^(EVENT_QUERY ^(IDENTIFIER[$EVENT_IDENTIFIER] ID) ^(QUERY qualifiedTerm) querySupplement)
  | EVENT_IDENTIFIER ':' qualifiedTerm querySupplement ->  ^(EVENT_QUERY ^(IDENTIFIER[$EVENT_IDENTIFIER] ID["eid-" + virtualIdentifier++]) ^(QUERY qualifiedTerm) querySupplement)
  | STATE_IDENTIFIER ID ':' qualifiedTerm querySupplement ->  ^(STATE_QUERY ^(IDENTIFIER[$STATE_IDENTIFIER] ID) ^(QUERY qualifiedTerm) querySupplement) 
  | STATE_IDENTIFIER ':' qualifiedTerm querySupplement ->  ^(STATE_QUERY ^(IDENTIFIER[$STATE_IDENTIFIER] ID["sid-" + virtualIdentifier++]) ^(QUERY qualifiedTerm) querySupplement) 
  ;

action
  : CONCURRENT '{' atomicAction (','? atomicAction)* ','? '}' ->  ^(COMPOUND atomicAction+)
  | COMPOUND '{' atomicAction (','? atomicAction)* ','? '}'   ->  ^(COMPOUND atomicAction+)
  | atomicAction                                              ->  ^(COMPOUND atomicAction+)
  ;

/** constraint: action term is actually defined */  
atomicAction
  : ACTION_IDENTIFIER ID? ':' qualifiedTerm ->  ^(ACTION_EXECUTION qualifiedTerm)
  ;


querySupplement
  : (supplement+=where | supplement+=let | supplement+=grouping)*
  ->  ^(SUPPLEMENT $supplement*)
  ;

negationSupplement
  : where?
  ->  ^(SUPPLEMENT where?)
  ;

actionSupplement
  : where*           // where is too generic here, only temporal relation can be stated in the where part of an expression. var X - var Y <= 0 may be kind of a temporal relation for event queries, but not for action definitions
  ->  ^(SUPPLEMENT where*)
  ;

typeSupplement
  : tupleConstraints?
  -> ^(SUPPLEMENT tupleConstraints?)
  ;

schemaSupplement
  : tupleConstraints? relationConstraints? conflictResolution?
  ->  ^(SUPPLEMENT tupleConstraints? relationConstraints? conflictResolution?)
  ;

let
  : LET '{' unification (','? unification)* ','? '}'
  ->  ^(LET unification+)
  ;

unification
  : dataVariable '=' expr
  -> ^(UNIFICATION dataVariable expr)
  | dataVariable '=' relativeTimerOp '(' timeInterval ',' durationValue ')'  
  -> ^(UNIFICATION dataVariable ^(relativeTimerOp timeInterval durationValue))
  ;

grouping
  : 'group by' '{' binding (',' binding)* '}'
  ->  ^(AGGREGATION ^(GROUPING binding+) AGGREGATION)
  
  | 'group by' '{' binding (',' binding)* '}' 'aggregate' '{' aggregation (',' aggregation)* '}'
  ->  ^(AGGREGATION ^(GROUPING binding+) ^(AGGREGATION aggregation+))
  ;

binding
  : dataVariable
  | identifier
  ;

aggregation
  : dataVariable '=' AGGREGATION_OP '(' binding ')' -> ^(UNIFICATION dataVariable ^(AGGREGATION_OP binding))
  | dataVariable '=' MAX '(' binding ')' -> ^(UNIFICATION dataVariable ^(AGGREGATION_OP[$MAX] binding))
  | dataVariable '=' MIN '(' binding ')' -> ^(UNIFICATION dataVariable ^(AGGREGATION_OP[$MIN] binding))  
  ;

/** constraint: no variables occur in mathFormula */    
tupleConstraints
  : 'where' '{' atomicMathFormula (',' atomicMathFormula)* '}'
  ->  ^(TUPLE_CONSTRAINTS atomicMathFormula+)
  ;

relationConstraints
  : 'constraints' '{' atomicRelationConstraint (',' atomicRelationConstraint)* '}' 
  -> ^(RELATION_CONSTRAINTS atomicRelationConstraint+)
  ;

conflictResolution
	: 'on conflict select' '{' atomicSelection (',' atomicSelection)* '}'
	-> ^(CONFLICT_RESOLUTION atomicSelection+)
	;

atomicSelection
	: MIN path	-> ^(SELECT[$MIN] path)
	| MAX path	-> ^(SELECT[$MAX] path)	
	;

atomicRelationConstraint
  : CONSTRAINT '(' path ')'                     ->  ^(CONSTRAINT path)
  | path CONSTRAINT path '{' path '}'           ->  ^(CONSTRAINT path+) 
  | CONSTRAINT '(' path '=' clockFunction ')'   ->  ^(CONSTRAINT ^(UNIFICATION path clockFunction))
  ;

where
  : CONDITION CONJUNCTION '{' conditionFormula (',' conditionFormula)*  '}'  -> ^(CONDITION ^(CONJUNCTION conditionFormula+))
  | CONDITION DISJUNCTION '{' conditionFormula (',' conditionFormula)*  '}'  -> ^(CONDITION ^(DISJUNCTION conditionFormula+))  
  | CONDITION NEGATION '{' conditionFormula (',' conditionFormula)*  '}'  	 -> ^(CONDITION ^(NEGATION conditionFormula+))  
  | CONDITION '{' conditionFormula (',' conditionFormula)*  '}'        			 -> ^(CONDITION ^(CONJUNCTION conditionFormula+))
  ;

conditionFormula
  : CONJUNCTION '{' conditionFormula (',' conditionFormula)* '}'	->  ^(CONJUNCTION conditionFormula+)
  | DISJUNCTION '{'conditionFormula (',' conditionFormula)* '}'   ->  ^(DISJUNCTION conditionFormula+)
  | NEGATION '{'conditionFormula (',' conditionFormula)* '}'   		->  ^(NEGATION conditionFormula+)
  | atomicConditionFormula
  ;
  
atomicConditionFormula
  :  atomicIntervalFormula
  |	 atomicMathFormula
  |  BOOLEAN
  ;

/** constraints: 
  *   either both or none of the expressions is of type duration 
  */
atomicMathFormula 
	:	expr arithmeticRelation expr ->  ^(arithmeticRelation expr expr)
	;

atomicIntervalFormula
  :	'{' timeInterval ',' timeInterval 
  	(	                    '}' op='apart-by' durationValue
  	|	(',' timeInterval)* '}' op='within' durationValue
  	)
  ->  ^(TIME_RELATION[$op] ^(IDENTIFIERS timeInterval+) durationValue)
  
  | timeInterval
  	(	intervalRelation timeInterval	->	^(intervalRelation timeInterval timeInterval)
  	|	timepointRelation timePoint	  ->	^(timepointRelation timeInterval timePoint)		
  	)
  ;

/** constraint: variable is of type timeinterval */
timeInterval
  : dataVariable
  |	path
  | identifier
  | relativeTimerOp '(' timeInterval ',' durationValue ')'  ->  ^(relativeTimerOp timeInterval durationValue)
  ;
  
durationValue
	:	duration
	| constant
	;
  
/** constraint: variable is of type timepoint */
timePoint
  : dataVariable
  | path
  | intervalOp '(' timeInterval ')'  ->  ^(intervalOp timeInterval) 
  ;

/** constraint: all timeunits are are pairwise distinct */
duration
  : time+
  ->  ^(DURATION time+)
  ;

time
  : NUMBER TIME_UNIT  ->  ^(TIME_UNIT NUMBER)
  | NUMBER MIN        ->  ^(TIME_UNIT[$MIN] NUMBER)  
  ;

/* XCerpt- */

/** 
 * constraints: 
 *   all term labels are pairwise distinct
 *   no cyclic definitions
 */
term
  : expr                           ->  expr
  | label '{' '}'                  ->  label
  | label '{' term (',' term)* '}' ->  ^(label term+)
  ;

qualifiedTerm
	:	expr                          ->  expr
	| path '{' '}'									->	path
	| path '{' term (',' term)* '}'	->	^(path term+)
	;

/** constraints: 
 *    variables are unified with basic types
 *    constants are unified with basic types
 *    identifiers are unified with identifier types
 */ 
termLeaf
  : dataVariable
  | clockFunction
  | identifier
  | constant
  | duration
  | path
  | BOOLEAN
  | STRING
  | NUMBER
  ;

/**
 * constraints:
 *   all constTerm labels are pairwise distinct
 *   no cyclic definitions
 */
constTerm
  : STRING
  | NUMBER
  | BOOLEAN
  | constant
  | duration
  | label '{' '}'                             ->  label
  | label '{' constTerm (',' constTerm)* '}'  ->  ^(label constTerm+)
  ;


/** 
 * constraints: 
 *   all schemaDefinition labels are pairwise distinct
 *   no cyclic definitions
 */
schemaDefinition
  : label '{' '}'                                           ->  label
  | label '{' basicType '}'                                 ->  ^(label basicType)
  | label '{' compositeType '}'                             ->  ^(label compositeType)
  | label '{' schemaDefinition (',' schemaDefinition)* '}'  ->  ^(label schemaDefinition+)
  ;
  
qualifiedSchemaDefinition
  : path '{' '}'                                           ->  path
  | path '{' basicType '}'                                 ->  ^(path basicType)
  | path '{' compositeType '}'                             ->  ^(path compositeType)
  | path '{' schemaDefinition (',' schemaDefinition)* '}'  ->  ^(path schemaDefinition+)
  ;

/** constraint: no duplicate specification of modifiers */
modifiers
  : modifier*
  ->  ^(MODIFIERS modifier*)
  ;

modifier
  : externalModifier
  | visibilityModifier
	| dynamabilityModifier
  ;

externalModifier
	: EXTERNAL_MODIFIER
	->  ^({token($externalModifier.text, MODIFIER, $externalModifier.start)})
  ;
  
visibilityModifier
  : VISIBILITY_MODIFIER
  ->  ^({token($visibilityModifier.text, MODIFIER, $visibilityModifier.start)})
  ;
  
dynamabilityModifier
  : DYNAMABILITY_MODIFIER
  ->  ^({token($dynamabilityModifier.text, MODIFIER, $dynamabilityModifier.start)})
  ;

expr  
  : mathExpr
  | identifier
  ;
    
mathExpr
  : (multExpr -> multExpr)
    ( (op='+' | op='-') multExpr
      ->  ^(ARITHMETIC_OP[$op] $mathExpr multExpr)
    )*
  ;
  
multExpr
  : (powExpr -> powExpr)
    ( (op='*' | op='/') powExpr
      ->  ^(ARITHMETIC_OP[$op] $multExpr powExpr)
    )*
  ;

powExpr
  : (atom -> atom)
    ( op='^' atom
      ->  ^(ARITHMETIC_OP[$op] $powExpr atom)
    )*
  ;

/**
 * constraints:
 *   constants used in aritmetic expressions need to be a number
 */
atom
  : '(' mathExpr ')'                                               ->  mathExpr
  | (op='least' | op='greatest') '(' mathExpr (',' mathExpr)* ')'  ->  ^(ARITHMETIC_OP[$op] mathExpr+)
  | op='-' atom                                                    ->  ^(ARITHMETIC_OP[$op] atom)
  | op='id' '(' (path | identifier) ')'                            ->  ^(IDENTIFIER_OP[$op] path? identifier?)  
  | intervalOp '(' timeInterval ')'	                               ->  ^(intervalOp timeInterval)
  | basicType '(' mathExpr ')' 												             ->  ^(CAST_OP[$basicType.text] mathExpr)
  | clockFunction
  | dataVariable
  | constant
  | path
  | duration
  | BOOLEAN
  | NUMBER
  | STRING
  ;

path
  : label ('.' label)*
  ->  ^({token($path.text, PATH, $path.start)})
  ;

clockFunction
  : 'system' '.' (fn='now' | fn='tuple_id_seq') '(' ')'  ->  CLOCK_FUNCTION[$fn]
  ;

label
  : AGGREGATION_OP 
  ->  TERM[$label.text] 
  | intervalOp        
  ->  TERM[$label.text]
  | ('action' | 'event' | 'state' | ACTION_SUFFIX | STATE_SUFFIX)
  ->  TERM[$label.text]
  | ('DESCRIPTION' | 'TYPE' | 'EVENT' | 'ACTION' | 'END')
  ->  TERM[$label.text] 
  | IDENTIFIER_OP
  ->  TERM[$label.text]
  | TIME_UNIT
  ->  TERM[$label.text]   
  | ID
  ->  TERM[$label.text]
  | ID d='$' ACTION_SUFFIX
  -> TERM[$label.text]
  | ID d='$' STATE_SUFFIX ('$' ACTION_SUFFIX)?
   -> TERM[$label.text]
  ;


type
  : basicType
  | compositeType
  ;

basicType 
  : (t='string' | t='int' | t='integer' | t='long' | t='double' | t='float' | t='boolean' | t='identifier' | t='timestamp' | t='duration')  
  ->  BASIC_TYPE[$t]
  ;

/** constraint: composite type ID is actually defined */
compositeType 
  : path
  ->  ^({token($path.text, COMPOSITE_TYPE, $path.start)})
  ;

identifier
  : (i=EVENT_IDENTIFIER | i=STATE_IDENTIFIER | i=ACTION_IDENTIFIER) ID
  ->  ^(IDENTIFIER[$i] ID)
  ;

dataVariable
  : VARIABLE ID
  ->  ^(VARIABLE ID)
  ;

/** constraint: constant ID is actually defined */
constant
  : CONSTANT path
  ->  ^(CONSTANT path)
  ;

/* this are actually tokens */

intervalRelation
  : (rel='before' | rel='contains' | rel='overlaps' | rel='after' | rel='during' | rel='overlapped-by'
      | rel='starts' | rel='finishes' | rel='meets' | rel='started-by' | rel='finished-by'
      | rel='met-by' | rel='equals' | rel='while' | rel='valid-during')
  ->  TIME_RELATION[$rel]
  ;

timepointRelation
	:	(rel='valid-at'	| rel='valid-at-distance') -> TIME_RELATION[$rel]
	;

relativeTimerOp 
  : (op='extend' | op='shorten' | op='extend-begin' | op='shorten-begin' | op='shift-forward'
      | op='shift-backward' | op='from-end' | op='from-end-backward' | op='from-start'
      | op='from-start-backward' | op='from-begin' |  op='from-begin-backward')
  ->  RELATIVE_TIMER_OP[$op]
  ;

intervalOp 
  : (op='begin' | op='end')
  ->  INTERVAL_OP[$op]
  ;

arithmeticRelation 
  : (rel='<' | rel='<=' | rel='=' | rel='!=' | rel='>' | rel='>=')
  ->  ARITHMETIC_RELATION[$rel]
  ;