package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.antlr.runtime.CommonToken;
import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


/**
 * This transformation does several things in regard to aggregation in event specification heads:
 * 1. Groupings inside of the heads are moved to the end of the body.
 * 2. Aggregation operations (like max(var A), min(var A), avg(var B), etc.) are replaced by variables.
 *    The operations are then moved to the aggregation portion of the last grouping (relative to the
 *    occurrence of the aggregation operation) and in there the replacement variables are unified with
 *    their respective aggregation operations.
 *    
 * This transformation could be split up into two separate transformations. However both steps can be
 * implemented more efficiently if combined.
 * 
 * @author scherr
 *
 */
public class AggregationAnywhere extends Transformation {
	// Only basic types allowed.
	// Only map those operators that have a definite return type (independent of parameter type).
	private HashMap<String, String> opNameToTypeNameMap = new HashMap<String, String>() {
		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;

		{
			// ... or float?
			put("avg", "double");
			put("mean", "double");
			put("count", "long");
		}
	};
	
	@Override
	public void apply(CommonTree tree) {
		for (CommonTree eventSpecification : collectEventSpecificationNodes(tree)) {
			String temp = null;
			if (EMIT_ANNOTATIONS) {
				temp = eventSpecification.toStringTree();
			}
			
			
			CommonTree head = (CommonTree) eventSpecification.getChild(0);
			
			assert(head.getType() == DuraParser.HEAD);
			
			CommonTree body = (CommonTree) eventSpecification.getChild(1);
			CommonTree composition = (CommonTree) body.getChild(0);
			CommonTree supplement = (CommonTree) composition.getChild(composition.getChildCount() - 1);
				
			boolean aggOpsInLetOrWhereExisted = handleAggOpsInLetOrWhere(body, composition, supplement);
			boolean aggOpsInInHeadExisted = handleAggOpsInHead(head, body, supplement);
			
			if (EMIT_ANNOTATIONS) {
				if (aggOpsInLetOrWhereExisted || aggOpsInInHeadExisted) {
					CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
					
					if (aggOpsInLetOrWhereExisted) {
						annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Handled aggregation operators in let or where blocks in deductive rule " + temp + ".\"")));
					}
					
					if (aggOpsInInHeadExisted) {
						annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Handled aggregation operators in head of deductive rule " + temp + ".\"")));
					}
					
					
					eventSpecification.getParent().addChild(annotation);
				}
			}
		}
			
		resetNextNameId();
	}
	
	private boolean handleAggOpsInLetOrWhere(CommonTree body, CommonTree composition, CommonTree supplement) {
		if (supplement.getChildCount() > 0) {
			CommonTree lastAggregation = null;
			for (Object c : supplement.getChildren()) {
				CommonTree suppChild = (CommonTree) c;
				
				if (suppChild.getType() == DuraParser.AGGREGATION) {
					lastAggregation = suppChild;
				} else {
					List<CommonTree> aggregationOpNodes = collectAggregationOpNodes(suppChild);
					if (!aggregationOpNodes.isEmpty()) {
						if (lastAggregation == null) {
							throw new RuntimeException("Aggregation in supplement without grouping.");
						}
						
						HashMap<String, CommonTree> treeStrToVarMap = new HashMap<String, CommonTree>();
						for (CommonTree aggregationOp : aggregationOpNodes) {
							// TODO: Make this variable usage optimization optional.
							String treeStr = aggregationOp.toStringTree();
							CommonTree variable = treeStrToVarMap.get(treeStr);
							if (variable == null) {
								variable = createVariableNode();
								treeStrToVarMap.put(treeStr, variable);
								
								int index = aggregationOp.getChildIndex();
								aggregationOp.getParent().replaceChildren(index, index, variable);
								
								CommonTree variableCopy = copyTree(variable);
													
								CommonToken typeToken = determineAndGetTypeToken(body, aggregationOp, suppChild);
								
								CommonTree unification = createUnificationNode(typeToken, variableCopy, aggregationOp);
								lastAggregation.addChild(unification);
							} else {
								variable = new CommonTree(variable);
								
								int index = aggregationOp.getChildIndex();
								aggregationOp.getParent().replaceChildren(index, index, variable);
							}
						}
					}
				}
			}
			
			if (lastAggregation != null) {
				return true;
			}
		}
		
		return false;
	}
	
	private boolean handleAggOpsInHead(CommonTree head, CommonTree body, CommonTree supplement) {	
		CommonTree aggregation = null;
		if (head.getChildCount() > 1) {
			aggregation = (CommonTree) head.deleteChild(1);
		}
		
		if (aggregation != null) {
			// Move grouping from head to body...
			supplement.addChild(aggregation);
		} else {
			if (supplement.getChildCount() > 0) {
				CommonTree maybeAggInSupp = (CommonTree) supplement.getChild(supplement.getChildCount() - 1);
			
				if (maybeAggInSupp.getType() == DuraParser.AGGREGATION) {
					aggregation = maybeAggInSupp;
				}
			}
		}
		
		// Move aggregation operations from head term to aggregation (now) in body.
		CommonTree headTerm = (CommonTree) head.getChild(0);
		List<CommonTree> aggregationOpNodes = collectAggregationOpNodes(headTerm);
		if (!aggregationOpNodes.isEmpty()) {
			if (aggregation == null) {
				throw new RuntimeException("Aggregation in rule head without grouping.");
			}
		
			HashMap<String, CommonTree> treeStrToVarMap = new HashMap<String, CommonTree>();
			for (CommonTree aggregationOp : aggregationOpNodes) {
				// TODO: Make this variable usage optimization optional.
				String treeStr = aggregationOp.toStringTree();
				CommonTree variable = treeStrToVarMap.get(treeStr);
				if (variable == null) {
					variable = createVariableNode();
					treeStrToVarMap.put(treeStr, variable);
					
					int index = aggregationOp.getChildIndex();
					aggregationOp.getParent().replaceChildren(index, index, variable);
					
					CommonTree variableCopy = copyTree(variable);
										
					CommonToken typeToken = determineAndGetTypeToken(body, aggregationOp, aggregation);
					
					CommonTree unification = createUnificationNode(typeToken, variableCopy, aggregationOp);
					aggregation.addChild(unification);
				} else {
					variable = new CommonTree(variable);
					
					int index = aggregationOp.getChildIndex();
					aggregationOp.getParent().replaceChildren(index, index, variable);
				}
			}
		}
		
		if (aggregation != null) {
			return true;
		}
		
		return false;
	}
	
	/**
	 * This method collects all aggregation operator nodes (^(AGGREGATION_OP[...] variable)) by breadth
	 * first search. It is mainly supposed to be called on (head) terms, arithmetic expression, etc.
	 * 
	 * @param tree
	 * @return
	 */
	protected List<CommonTree> collectAggregationOpNodes(CommonTree tree) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(tree);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.AGGREGATION_OP) {
				nodes.add(node);
				continue;
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
		return nodes;
	}
	
	// TODO: This needs a more efficient implementation
	private CommonToken determineAndGetTypeToken(CommonTree body, CommonTree aggregationOp, CommonTree beforeSupplementChild) {
		// Some types are (constantly) determined by the aggregation function.
		String typeName = opNameToTypeNameMap.get(aggregationOp.getText());
		if (typeName != null) {
			return new CommonToken(DuraParser.BASIC_TYPE, typeName);
		}
		
		String variableName = aggregationOp.getChild(0).getChild(0).getText();
		
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(body);
		
		CommonTree tempTerm = null;
		
		int beforeSuppChildIndex = beforeSupplementChild.getChildIndex();
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.VARIABLE && node.getParent().getType() == DuraParser.TERM) {
				// We are only interested in untyped variables if they occur in query terms.
				
				if (node.getChild(0).getText().equals(variableName)) {
					tempTerm = node;
					break;
				}
			} else if (node.getType() == DuraParser.TYPED_VARIABLE) {
				if (node.getChild(1).getText().equals(variableName)) {
					return (CommonToken) copyTree(((CommonTree) node.getChild(0))).getToken();
				}
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					
					CommonTree parent = (CommonTree) node.getParent();
					if (node.getType() == DuraParser.SUPPLEMENT &&
							(parent.getType() == DuraParser.CONJUNCTION || parent.getType() == DuraParser.DISJUNCTION) &&
							child.getChildIndex() >= beforeSuppChildIndex) {
						break;
					}
					
					nodeQueue.add(child);
				}
			}
		}
		
		if (tempTerm == null) {
			throw new RuntimeException("Type matching failed (variable: " + variableName + ")!" );
		}
		
		ArrayList<String> pathToTermRoot = new ArrayList<String>();
		
		CommonTree parent = (CommonTree) tempTerm.getParent();
		while(parent.getType() == DuraParser.TERM) {
			pathToTermRoot.add(parent.getText());
			tempTerm = parent;
			parent = (CommonTree) tempTerm.getParent();
		}
		
		String termLabel = tempTerm.getText();
		
		int schemaType = (parent.getParent().getType() == DuraParser.EVENT_QUERY ? DuraParser.EVENT : DuraParser.STATEFUL_OBJECT);
		
		if (schemaType == DuraParser.EVENT) {
			tempTerm = eventSchemataMap.get(termLabel);
		} else {
			tempTerm = stateSchemataMap.get(termLabel);
		}
		
		if (tempTerm == null) {
			throw new RuntimeException("Type matching failed (variable: " + variableName + ")!" );
		}
		
		// tempTerm is the schema term
		for (int i = pathToTermRoot.size() - 2; i >= 0; i--) {
			String label = pathToTermRoot.get(i);
			
			if (tempTerm.getChildCount() > 0) {
				for (Object c : tempTerm.getChildren()) {
					CommonTree schemaChild = (CommonTree) c;
					String schemaChildLabel = schemaChild.getText();
				
					if (label.equals(schemaChildLabel)) {
						tempTerm = schemaChild;
						continue;
					}
				}
			} else {
				break;
			}
		}
		if (tempTerm.getChildCount() > 0) {
			tempTerm = (CommonTree) tempTerm.getChild(0);
		}
		
		if (tempTerm.getType() == DuraParser.BASIC_TYPE || tempTerm.getType() == DuraParser.COMPOSITE_TYPE) {
			return (CommonToken) copyTree(tempTerm).getToken();
		} else {
			throw new RuntimeException("Type matching failed (variable: " + variableName + ")!" );
		}
	}
}
