package de.lmu.ifi.pms.cep.dura.compiler.durac.collections;

import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.ACTION;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.EVENT_QUERY;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.IDENTIFIER;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.STATE_QUERY;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.TERM;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.VARIABLE;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.actionIdentifierPrefix;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.eventIdentifierPrefix;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.eventIntervalName;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.identifierAttribute;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.stateIdentifierPrefix;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Set;
import java.util.Stack;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.durac.DuraCQueryCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.durac.tsa.TimeInterval;
import de.lmu.ifi.pms.cep.dura.compiler.durac.tsa.UntypedIdentifier;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraAmbiguousMappingException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCompilationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraParseException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnknownVariableException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.UnexpectedTokenException;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.IdentifierAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.NamedTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.UntypedVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.CompositeType;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchemaTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;

/**
 * QueryVariables is a mapping from dura variable names to (possibly multiple)
 * corresponding attributes of the event stream.
 * 
 * @author hausmann
 * 
 */
public class QueryVariables  {
	private static long ids = 0;
	
	protected HashMap<Variable<?>, ArrayList<Term<? extends Type>>> mapping = new HashMap<Variable<?>, ArrayList<Term<? extends Type>>>();

	static {
		assert(BasicTypeID.INT.compareTo(BasicTypeID.LONG) < 0);
		assert(BasicTypeID.LONG.compareTo(BasicTypeID.FLOAT) < 0);
		assert(BasicTypeID.FLOAT.compareTo(BasicTypeID.DOUBLE) < 0);
	}

	protected QueryVariables() {}
	
	
	public QueryVariables(final Variable<?> variable, final Term<? extends Type> term) {
		this.addMapping(variable, term);
	}
	
	
	public QueryVariables(final DuraTree query, final DuraSchema schema, final Preamble preamble) {
		assert(query.getType() == EVENT_QUERY ||
				query.getType() == STATE_QUERY ||
				query.getType() == ACTION);
		
		HashMap<Variable<?>, ArrayList<String>> untypedMapping = new HashMap<Variable<?>, ArrayList<String>>();
		
		DuraTree identifier = query.getChild(0);
		DuraTree queryTerm = query.getChild(1).getChild(0);

		this.collectIdentifier(identifier);
		
		//collect variable/attribute mappings
		for (int i=0; i<queryTerm.getChildCount(); i++) {
			DuraTree child = queryTerm.getChild(i);
			
			this.collectQueryTermVariables(child, untypedMapping, new Stack<String>());
		}

		//type collected attributes
		try {
			TemporalStreamSchema streamSchema = schema.streams.getStreamByName(TemporalStreamSchemaTypeID.EVENT, preamble, queryTerm.getText()).schema;
			
			for (Variable<?> var : untypedMapping.keySet()) {
				ArrayList<String> untypedAttributes = untypedMapping.get(var);
				ArrayList<Term<? extends Type>> typedAttributes = new ArrayList<Term<? extends Type>>(untypedAttributes.size());
				
				for (String attribute : untypedAttributes) {
					typedAttributes.add(TypedAttribute.create(TSA_SyntaxPatterns.prefixAttribute(identifier.getChild(0).getText(), attribute), streamSchema.typeOf(attribute)));
				}
				
				this.mapping.put(var, typedAttributes);
			}
			
		} catch (DuraUnspecificCompliationException e) {
			throw new DuraCompilationException(e, query);
		} catch (TSA_TypeException e) {
			throw new DuraCompilationException(e, query);
		}
	}
	
	
	private void collectIdentifier(final DuraTree identifier) {
		assert(identifier.getType() == IDENTIFIER);
		assert(identifier.getText().equals(eventIdentifierPrefix));
		
		String identifierName = identifier.getChild(0).getText();
		
		BasicTypeVariable identifierVariable = (BasicTypeVariable) variableFromToken(identifier);
		NamedTypeVariable<CompositeType> intervalVariable = NamedTypeVariable.create(identifierVariable.name, TypeDefinitions.TIME_INTERVAL);
		
		this.addMapping(identifierVariable, IdentifierAttribute.create(TSA_SyntaxPatterns.prefixAttribute(identifierName, identifierAttribute)));
		this.addMapping(intervalVariable, new TimeInterval(Term.attr(TSA_SyntaxPatterns.prefixAttribute(identifierName, eventIntervalName), TypeDefinitions.TIME_INTERVAL)));
	}
	
	
	private void collectQueryTermVariables(final DuraTree term, final HashMap<Variable<?>, ArrayList<String>> mapping, final Stack<String> path) {
		if (term.getType() == VARIABLE) {
			throw new DuraParseException(String.format("expected an attribute name, found a variable: %s", term.toStringTree()), term.getPositionDescription());
		}
		
		assert(term.getType() == TERM);

		for (int i=0; i<term.getChildCount(); i++) {
			DuraTree child = term.getChild(i);
			
			path.push(term.getText());
			
			switch (child.getType()) {
			case TERM:
				this.collectQueryTermVariables(child, mapping, path);
				
				break;
				
			case VARIABLE:
				Variable<?> variable = variableFromToken(child);
				String attribute = DuraCQueryCompiler.joinPath(path);
				
				if (! mapping.containsKey(variable)) {
					mapping.put(variable, new ArrayList<String>());		
				}
				
				mapping.get(variable).add(attribute);

				break;
			}
			
			path.pop();
		}
	}
	
	
	public static UnambiguousQueryVariables generateImplicitAttribute(UntypedVariable variable, Type type) {
		TypedAttribute<? extends Type> attribute = TypedAttribute.create(String.format(".dura.implicit.var-%s", ids++), type);
		
		return new UnambiguousQueryVariables(variable, attribute);
	}
	
	
	private void addMapping(final Variable<?> variable, final Term<? extends Type> attribute) {
		if (! this.mapping.containsKey(variable)) {
			this.mapping.put(variable, new ArrayList<Term<? extends Type>>());
		}
		
		this.mapping.get(variable).add(attribute);
	}
	
	
	private void addMapping(final Variable<?> variable, final ArrayList<Term<? extends Type>> attributes) {
		if (! this.mapping.containsKey(variable)) {
			this.mapping.put(variable, new ArrayList<Term<? extends Type>>());
		}
		
		this.mapping.get(variable).addAll(attributes);
	}
	
	
	public static Variable<?> variableFromToken(final DuraTree variable) {
		switch (variable.getType()) {
		case VARIABLE:
			return Term.var(variable.getChild(0).getText());

		/*
		case TYPED_VARIABLE:
			BasicTypeID type = DuraBasicTypesMapping.fromDuraName(variable.getChild(0).getText());
			
			return BasicTypeVariable.create(variable.getChild(1).getText(), type);
		*/
			
		case IDENTIFIER:
			return  BasicTypeVariable.create(String.format("%s %s", variable.getText(), variable.getChild(0).getText()), BasicTypeID.IDENTIFIER);

		default:
			throw new UnexpectedTokenException(variable);
		}
	}

	
	public static QueryVariables unionUnambiguousMappings(final UnambiguousQueryVariables... variables) {
		return unionUnambiguousMappings(Arrays.asList(variables));
	}
	
	
	public static QueryVariables unionUnambiguousMappings(final Collection<UnambiguousQueryVariables> variables) {
		QueryVariables result = new QueryVariables();
		
		for(QueryVariables mapping : variables) {
			for (Variable<?> var : mapping.mapping.keySet()) {
				if (! result.mapping.containsKey(var)) {
					result.mapping.put(var, new ArrayList<Term<? extends Type>>());
				}
				
				result.mapping.get(var).add(mapping.mapping.get(var).get(0));
			}
		}
		
		return result;
	}
	
	
	public QueryVariables dropMappings(String... variableNames) {
		QueryVariables result = new QueryVariables();
		
		outer: for (Variable<?> var : this.mapping.keySet()) {
			for (String name : variableNames) {
				if (var.name.equals(name)) {
					continue outer;
				}
			}

			result.addMapping(var, this.mapping.get(var));
		}
		
		return result;
	}
	
	
	/**
	 * This function returns a ConditionFormula that ensures that the values of
	 * all attributes of the same dura variable are actually equal.
	 * 
	 * When the ConditionFormula is applied to a variable mapping, one can pick
	 * an arbitrary attribute to represent a dura variable.
	 * 
	 * @return
	 */
	public ConditionFormula compileImplicitVariableUnifications() {
		ArrayList<ConditionFormula> conditions = new ArrayList<ConditionFormula>();
		
		for (ArrayList<Term<? extends Type>> attribute : this.mapping.values()) {
			for (int i=0; i<attribute.size(); i++) {
				for (int j=i+1; j<attribute.size(); j++) {
					conditions.add(new BinaryComparissonFormula(BinaryComparissonOperatorID.EQUAL, attribute.get(i), attribute.get(j)));
				}
			}
		}
		
		if (conditions.size() == 0) {
			return ConditionFormula.Top;
		} else {
			return new ConjunctiveConditionFormula(new FixedArray<ConditionFormula>(conditions));
		}
	}
	
	
	public UnambiguousQueryVariables dropAmbiguousMappings() {
		try {
			return convertToUnambiguousMapping(false);
		} catch (DuraAmbiguousMappingException e) {
			throw new InternalError();		//impossible case
		}
	}

	
	public UnambiguousQueryVariables convertToUnambiguousMapping() throws DuraAmbiguousMappingException {
		return convertToUnambiguousMapping(true);
	}
	
	
	private UnambiguousQueryVariables convertToUnambiguousMapping(boolean exceptionOnInformationLoss) throws DuraAmbiguousMappingException {
		UnambiguousQueryVariables result = new UnambiguousQueryVariables();
		
		for (Variable<?> var : this.mapping.keySet()) {
			ArrayList<Term<? extends Type>> mappings = this.mapping.get(var);
			
			if (exceptionOnInformationLoss && mappings.size()>1) {
				throw new DuraAmbiguousMappingException(String.format("found ambiguous mapping for variable %s: %s", var.name, mappings.toString()));
			}
			
			result.addMapping(var, this.mapping.get(var).get(0));
		}
		
		return result;
	}
	
	
	public boolean containsVariable(Variable<?> variable) {
		try {
			getVariableAttributes(variable);
			
			return true;
		} catch (DuraUnknownVariableException e) {
			return false;
		}
	} 
		
	public ArrayList<Term<? extends Type>> getVariableAttributes(Variable<?> variable) throws DuraUnknownVariableException {
		if (variable instanceof UntypedIdentifier) {
			UntypedIdentifier identifier = (UntypedIdentifier) variable;
			
			NamedTypeVariable<CompositeType> event = NamedTypeVariable.create(String.format("%s %s", eventIdentifierPrefix, identifier.name), TypeDefinitions.TIME_INTERVAL); 
			NamedTypeVariable<CompositeType> state = NamedTypeVariable.create(String.format("%s %s", stateIdentifierPrefix, identifier.name), TypeDefinitions.TIME_INTERVAL); 
			NamedTypeVariable<CompositeType> action = NamedTypeVariable.create(String.format("%s %s", actionIdentifierPrefix, identifier.name), TypeDefinitions.TIME_INTERVAL); 
				
			if (this.mapping.containsKey(event)) {
				return this.mapping.get(event); 
			} else if (this.mapping.containsKey(state)) {
				return this.mapping.get(state); 
			} else if (this.mapping.containsKey(action)) {
				return this.mapping.get(action); 
			} else {
				throw new DuraUnknownVariableException(identifier);
			}
		} else if (this.mapping.containsKey(variable)) {
			return this.mapping.get(variable); 
		} else {
			throw new DuraUnknownVariableException(variable);
		}
	}

	
	/**
	 * Filters variable mappings so that for each variable, one (of potentially
	 * multiple mappings) remains in the result.
	 * 
	 * All variables that have mappings which are either all matching or are all
	 * not matching <tt>identifiers</tt> are filtered so that only one arbitrary
	 * mapping remains.
	 * 
	 * All variables that have mappings where one is matching
	 * <tt>identifiers</tt> and another one is not matching, are filtered so
	 * that only one mapping not matching <tt>identifiers</tt> remains.
	 * 
	 */
	public QueryVariables filterAmbiguosMatchingMappings(Collection<String> identifiers) {
		QueryVariables result = new QueryVariables();
		
		for (Variable<?> var : this.mapping.keySet()) {
			//first collect only mappings that are not matching the identifers that should be filtered
			outer: for (Term<? extends Type> term : this.mapping.get(var)) {
				for (String attr : term.AttributeNames()) {
					if (! identifiers.contains(TSA_SyntaxPatterns.splitAttributeName(attr)[0])) {
						result.addMapping(var, term);
						
						continue outer;
					}
				}
			}

			//if no non-matching mapping is found, use the matching ones 
			if (!result.mapping.containsKey(var)) {
				result.addMapping(var, this.mapping.get(var));
			}
		}
		
		return result;
	}
	
	
	/**
	 * Drops mappings for that every attribute of the mapping matches prefix.
	 * 
	 * If a the attribute of a mapping consists of several attributes of which
	 * some are matching the prefix and some are not, the mapping is kept.
	 */
	public QueryVariables dropMappingsMatchingPrefix(Collection<String> prefixes) {
		QueryVariables result = new UnambiguousQueryVariables();
		
		outer: for (Variable<?> var : this.mapping.keySet()) {
			Term<? extends Type> term = this.mapping.get(var).get(0);
			
			inner: for (String attribute : term.AttributeNames()) {
				for (String prefix : prefixes) {
					if (attribute.startsWith(prefix + ".")) {
						continue inner;
					}
				}
				
				result.addMapping(var, term);
				
				continue outer;
			}
		}
		
		return result;
	}
	
	
	public QueryVariables getMappingsMatchingPrefix(String prefix) {
		QueryVariables result = new UnambiguousQueryVariables();
		
		outer: for (Variable<?> var : this.mapping.keySet()) {
			Term<? extends Type> term = this.mapping.get(var).get(0);
			
			for (String attribute : term.AttributeNames()) {
				if (! attribute.startsWith(prefix + ".") && ! attribute.equals(prefix)) {
					continue outer;
				}
			}
			
			result.addMapping(var, term);
		}
		
		return result;
	}
	
	
	
	public Set<Variable<?>> getVariables() {
		return this.mapping.keySet();
	}
	
	
	public String getVariablesMatchingAttribute(String attribute) {
		Collection<String> pairs = new ArrayList<String>();
		
		for (Variable<?> var : this.mapping.keySet()) {
			for (Term<?> term : this.mapping.get(var)) {
				if (term instanceof Attribute) {
					if (((Attribute<?>) term).name.equals(attribute)) {
						pairs.add(String.format("%s/%s", attribute, var.toString()));
					}
				}
			}
		}
		
		String result = pairs.toString();
		
		return result.substring(1, result.length()-1);
	}
	
	
	@Override
	public String toString() {
		return this.mapping.keySet().toString();
	}
}