package de.lmu.ifi.pms.cep.dura.compiler.durac.formula_transformations;

import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.intervalBeginAttribute;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.intervalEndAttribute;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunction.apply;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.MINUS;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.PLUS;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.TIMESTAMP;

import java.util.HashMap;

import de.lmu.ifi.pms.cep.dura.compiler.durac.tsa.RelativeTimerOp;
import de.lmu.ifi.pms.cep.dura.compiler.durac.tsa.TimeInterval;
import de.lmu.ifi.pms.cep.dura.compiler.durac.tsa.TimeIntervalOp;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCExceptionWrapper;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.CompositeTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.MapReplaceAtomicTerm;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


class MapRewriteIntervalFunctions implements Adapter<Term<?>, Term<?>> {
	
	public static final MapRewriteIntervalFunctions instance = new MapRewriteIntervalFunctions();
	
	private static final HashMap<String, TimeInterval> relativeTimerFunctions= new HashMap<String, TimeInterval>();

	private static final TimestampAttribute i_begin = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute("i", intervalBeginAttribute));
	private static final TimestampAttribute i_end = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute("i", intervalEndAttribute));
	private static final DurationAttribute duration = DurationAttribute.create("duration");

	static {
		relativeTimerFunctions.put("extend", new TimeInterval(i_begin, apply(PLUS, TIMESTAMP.type, i_end, duration)));
		relativeTimerFunctions.put("extend-begin", new TimeInterval(apply(MINUS, TIMESTAMP.type, i_begin, duration), i_end));
		relativeTimerFunctions.put("shorten", new TimeInterval(i_begin, apply(MINUS, TIMESTAMP.type, i_end, duration)));
		relativeTimerFunctions.put("shorten-begin", new TimeInterval(apply(PLUS, TIMESTAMP.type, i_begin, duration), i_end));
		relativeTimerFunctions.put("shift-forward", new TimeInterval(apply(PLUS, TIMESTAMP.type, i_begin, duration), apply(PLUS, TIMESTAMP.type, i_end, duration)));
		relativeTimerFunctions.put("shift-backward", new TimeInterval(apply(MINUS, TIMESTAMP.type, i_begin, duration), apply(MINUS, TIMESTAMP.type, i_end, duration)));
		relativeTimerFunctions.put("from-end", new TimeInterval(i_end, apply(PLUS, TIMESTAMP.type, i_end, duration)));
		relativeTimerFunctions.put("from-end-backward", new TimeInterval(apply(MINUS, TIMESTAMP.type, i_end, duration), i_end));
		relativeTimerFunctions.put("from-start", new TimeInterval(i_begin, apply(PLUS, TIMESTAMP.type, i_begin, duration)));
		relativeTimerFunctions.put("from-start-backward", new TimeInterval(apply(MINUS, TIMESTAMP.type, i_begin, duration), i_begin));
	}

	@Override
	public Term<?> adapt(final Term<?> source) {
		try {
			if (source instanceof RelativeTimerOp) {
				return rewriteRelativeTimerFunctions(source);
			} else if (source instanceof TimeIntervalOp) {
				return rewriteTimepointFunctions(source);
			} else if (source instanceof CompositeTerm) {
				CompositeTerm<?> composite = (CompositeTerm<?>) source;
				
				Term<?>[] rewritenArguments = new Term[composite.arity()];
				
				for (int i=0; i<composite.arity(); i++) {
					rewritenArguments[i] = this.adapt(composite.arg(i));
				}
				
				return composite.newTerm(rewritenArguments);
			}
			
	 		return source;
		} catch (DuraUnspecificCompliationException e) {
			throw new DuraCExceptionWrapper(e);
		}
	}
	
	private static TimeInterval rewriteRelativeTimerFunctions(final Term<?> term) throws DuraUnspecificCompliationException {
		if (term instanceof RelativeTimerOp) {
			RelativeTimerOp function = (RelativeTimerOp) term;

			if (relativeTimerFunctions.containsKey(function.name)) {
				TimeInterval interval = rewriteRelativeTimerFunctions(function.arg(0));
				DurationValue dur = (DurationValue) function.arg(1);
				
				final FixedArrayNavigableMap<AtomicTerm<?>, Term<?>> mapping = new FixedArrayNavigableMap<AtomicTerm<?>, Term<?>>(
						new AtomicTerm[] { i_begin, i_end, duration},
						new Term[] {interval.begin(), interval.end(), dur }
				);
				
				return (TimeInterval) relativeTimerFunctions.get(function.name).replaceAtomicTerm(new MapReplaceAtomicTerm(mapping));
			} else {
				throw new DuraUnspecificCompliationException(String.format("found unknown function: %s", function.name));
			}
		} else {
			return MapRewriteTimeRelations.toTimeInterval(term);
		}
	}

	private static Term<?> rewriteTimepointFunctions(final Term<?> term) throws DuraUnspecificCompliationException {
		TimeIntervalOp function = (TimeIntervalOp) term;

		assert(function.arity() == 1);

		if (function.name.equals(intervalBeginAttribute)) {
			return rewriteRelativeTimerFunctions(function.arg(0)).begin();
		} else {
			assert(function.name.equals(intervalEndAttribute));
			
			return rewriteRelativeTimerFunctions(function.arg(0)).end();
		}
	}
}
