package de.lmu.ifi.pms.cep.event_mill.console;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.Version;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCompilationException;
import de.lmu.ifi.pms.cep.event_mill.EventMill;
import de.lmu.ifi.pms.cep.event_mill.configuration.Configuration;
import de.lmu.ifi.pms.cep.event_mill.configuration.DEFAULTS;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgram;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Translation;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.compatibility.HashMap;



public class MainThread extends AbstractConsoleThread {
	
	public static final String fileNameRegEx = "(?:[^ ,\\t\\n\\x0B\\f\\r]++(?: ++[^ ,\\t\\n\\x0B\\f\\r]++)*+)";
	
	public static final String compilerAliasRegEx = "[a-zA-Z_][a-zA-Z0-9_-]*";
	
	public static final String configCommandRegEx ="(?:config[ \\t]++(" + fileNameRegEx + ")[ \\t]*+)";
	
	public static final Pattern configCommandPattern = Pattern.compile(configCommandRegEx);
	
	public static final String commaSeparatorRegEx = "(?:[ \\t]*+,[ \\t]*+)";
	public static final String trailingLinearWhitespaceRegEx = "(?:[ \\t]*+)";
	
	public static final String compileCommandRegEx =	"(?:compile[ \\t]++((?:" + compilerAliasRegEx + "+[ \\t]++>[ \\t]++)*)" + TSA_SyntaxPatterns.ProgramFullNameRegEx + commaSeparatorRegEx +
														"((?:" + fileNameRegEx + ")?"+ 
														"(?:" + commaSeparatorRegEx + fileNameRegEx + ")*)" +
														trailingLinearWhitespaceRegEx + ")";
	public static final Pattern compileCommandPattern = Pattern.compile(compileCommandRegEx);
	
	public static final String loadSpecCommandRegEx = "(?:load spec[ \\t]++(" + fileNameRegEx + ")" + trailingLinearWhitespaceRegEx + ")";
	public static final Pattern loadSpecCommandPattern = Pattern.compile(loadSpecCommandRegEx);
	
	public static final String listSpecsCommandRegEx = "(?:list specs" + trailingLinearWhitespaceRegEx + ")";
	public static final String clearSpecsCommandRegEx = "(?:clear specs" + trailingLinearWhitespaceRegEx + ")";
	public static final String removeSpecCommandRegEx = "(?:remove spec[ \\t]++("+ TSA_SyntaxPatterns.ProgramFullNameRegEx +")" + trailingLinearWhitespaceRegEx + ")";
	public static final Pattern removeSpecCommandPattern = Pattern.compile(removeSpecCommandRegEx);
	public static final String initCommandRegEx = "(?:init[ \\t]+(" + TSA_SyntaxPatterns.program_instance_full_name_RegEx + ")" + commaSeparatorRegEx + 
	                                              "(" + SQL_Translation.program_schema_name_RegEx + ")" + commaSeparatorRegEx +
	                                              "(" + SQL_Translation.program_schema_name_RegEx + ")" + commaSeparatorRegEx +
	                                              "(" + SQL_Translation.program_schema_name_RegEx + ")" + commaSeparatorRegEx +
	                                              "(" + SQL_Translation.program_schema_name_RegEx + ")" + commaSeparatorRegEx +
	                                              "(" + SQL_Translation.program_schema_name_RegEx + ")" + 
	                                              trailingLinearWhitespaceRegEx + ")";
	public static final Pattern initCommandPattern = Pattern.compile(initCommandRegEx);
	public static final String listProgsCommandRegEx = "(?:list progs" + trailingLinearWhitespaceRegEx + ")";
	public static final String clearProgsCommandRegEx = "(?:clear progs" + trailingLinearWhitespaceRegEx + ")";
	public static final String removeInstCommandRegEx = "(?:remove prog[ \\t]++(" + TSA_SyntaxPatterns.program_instance_full_name_RegEx + ")" + trailingLinearWhitespaceRegEx + ")";
	public static final Pattern removeInstCommandPattern = Pattern.compile(removeInstCommandRegEx);
	public static final String runCommandRegEx = "(?:run[ \\t]++(" + TSA_SyntaxPatterns.program_instance_full_name_RegEx + ")" + trailingLinearWhitespaceRegEx + ")";
	public static final Pattern runCommandPattern = Pattern.compile(runCommandRegEx);
	
	public static final String clearCommandRegEx = "(?:clear" + trailingLinearWhitespaceRegEx + ")";
	public static final String uninstallCommandRegEx = "(?:uninstall" + trailingLinearWhitespaceRegEx + ")";
	
	EventMill eventMill;
	
	public MainThread(final Collection<Iterator<String>> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream, final Configuration configuration){
		this(new HashMap<String, String>(), immediateCommands, inputStream, outputStream);
	}
	
	public MainThread(final Map<String, String, ?> variables, final Collection<Iterator<String>> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream, final Configuration configuration){
		this(variables, immediateCommands, inputStream, outputStream);
		this.config(configuration);
	}
	
	public MainThread(final Collection<Iterator<String>> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream) {
		this(new HashMap<String, String>(), immediateCommands, inputStream, outputStream);
	}
	
	public MainThread(final Map<String, String, ?> variables, final Collection<Iterator<String>> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream) {
		super("MainThread", variables, immediateCommands, inputStream, outputStream);
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	void config(final Configuration configuration){
		if(this.eventMill == null){
			this.eventMill = new EventMill(configuration);
		}
		else{
			this.eventMill.config(configuration);
		}
	}
	
	@Override
	protected void afterSetVariable(final String var, final String value) {
		//FIXME more generic!!
		if(var.equals("source") & value != null){
			String source = this.getVariable("source");
			File file = new File(source);
			
			this.setVariable("source.dir", file.getParent());
			
			String result;					
			if(this.isSetVariable("result")){
				result = this.getVariable("result");
			}
			else{
				if(file.isDirectory()){
					result = file.getParentFile().getAbsolutePath() + "/" + file.getName() + ".xml";
				}
				else{
					result = (file.getAbsolutePath().endsWith(".dura") ? file.getAbsolutePath().substring(0, file.getAbsolutePath().length() - 5) : file.getAbsolutePath()) + ".xml";
				}
				this.setVariable("result", result);
			}
			
			String name;
			if(this.isSetVariable("name")){
				name = this.getVariable("name");
			}
			else{
				name = file.getName().replaceAll("[.]dura", "");
				this.setVariable("name", name);
			}
		}
			
		if((var.equals("name") | this.isSetVariable("name")) & value != null){
			String name = this.getVariable("name");
			String meta_schema;
			if(this.isSetVariable("meta-schema")){
				meta_schema = this.getVariable("meta-schema");
			}
			else{
				meta_schema = SQL_Translation.SQL_BasicTableName(name) + "_meta";
				this.setVariable("meta-schema", meta_schema);
			}
			String input_schema;
			if(this.isSetVariable("input-schema")){
				input_schema = this.getVariable("input-schema");
			}
			else{
				input_schema = SQL_Translation.SQL_BasicTableName(name) + "_input";
				this.setVariable("input-schema", input_schema);
			}
			String working_schema;
			if(this.isSetVariable("working-schema")){
				working_schema = this.getVariable("working-schema");
			}
			else{
				working_schema = SQL_Translation.SQL_BasicTableName(name) + "_working";
				this.setVariable("working-schema", working_schema);
			}
			String output_schema;
			if(this.isSetVariable("output-schema")){
				output_schema = this.getVariable("output-schema");
			}
			else{
				output_schema = SQL_Translation.SQL_BasicTableName(name) + "_output";
				this.setVariable("output-schema", output_schema);
			}
			String log_schema;
			if(this.isSetVariable("log-schema")){
				log_schema = this.getVariable("log-schema");
			}
			else{
				log_schema = SQL_Translation.SQL_BasicTableName(name) + "_log";
				this.setVariable("log-schema", log_schema);
			}
			
			String instance;
			if(this.isSetVariable("instance")){
				instance = this.getVariable("instance");
			}
			else{
				instance = "test";
				this.setVariable("instance", instance);
			}
		}
	}

	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public void run() {
		int exitCode = 0;
		try{
			this.info("Version " + Version.full());
			this.info("Starting");
			
			this.outputStream.println("\nEvent-Mill " + Version.full());
			this.outputStream.println("\nWelcome \n");			
		
			this.outputStream.println("For leaving type \"quit\"");
			
			if(this.eventMill == null || !this.eventMill.isConfigured()){
				this.info("Loading default configuration.");						
				try{
					if(Configuration.exists()){
						this.info("Default configuration exists. Loading default configuration.");
						
						try {
							this.config(Configuration.read());
							this.outputStream.println("Loaded default configuration");
							this.info("Loaded default configuration.");
						} catch (IOException e) {
							this.warn("Unable to load default configuration.", e);
						}
	
					}
					else{
						this.outputStream.println("Default configuration does not exist.");
						String defaultConfigFilePath = Configuration.defaultConfigFilePath();
						if(defaultConfigFilePath != null){
							this.outputStream.println("Expected at path: " + defaultConfigFilePath);	
						}
						else{
							this.outputStream.println("Expected on the classpath at: " + DEFAULTS.defaultConfigResourcePath);
						}
					}
				} catch (Exception e) {
					this.warn("Unable to load default configuration.", e);
					this.outputStream.println("Unable to load default configuration.");
				}
			}
			
			String command;
			boolean running = true;
			while(running && (this.eventMill == null || !this.eventMill.isConfigured())){
				this.outputStream.println("Proposed step: \"config\"");	
				try{
					command = this.nextCommand().trim();
				} catch (EndOfInputException e) {
					running = false; 
					this.info("End of input.",e); 
					break;
				}
				this.debug("Executing command: " + command);
				if(command.equals("")){
					continue;
				}
				else if(command.startsWith("set")){
					Matcher matcher = variableDefinitionPattern.matcher(command);
					
					if(matcher.matches()){
						String var = matcher.group(1);
						String value = matcher.group(2);
						this.setVariable(var, value);
					}
					else{
						this.outputStream.println("Illegal argument for set command.");
						this.skipRemainigCommands();
					}
				}
				else if("quit".equalsIgnoreCase(command)){
					running = false;
					
					this.debug("Stopping...");
				}
				else if(command.startsWith("config")){
					this.info("Starting configuration.");
					
					Matcher matcher = configCommandPattern.matcher(command);
					
					if(matcher.matches()){
						String configFile = matcher.group(1);
						this.info("Loading configuration.");						
						try{
							if(Configuration.existsFile(configFile)){
								this.info("Configuration exists. Loading configuration.");
								
								try {
									this.config(Configuration.readFile(configFile));
									this.outputStream.println("Loaded configuration");
									this.info("Loaded configuration.");
								} catch (IOException e) {
									this.outputStream.println("Unable to load configuration.");
									this.warn("Unable to load configuration.", e);
									this.skipRemainigCommands();
								}
								
							}
							else{
								this.outputStream.println("Configuration file does not exist: " + new File(configFile).getAbsolutePath());
								this.info("Configuration file does not exist: {}", new File(configFile).getAbsolutePath());
								this.skipRemainigCommands();
							}
						} catch (Exception e) {
							this.outputStream.println("Unable to load configuration.");
							this.warn("Unable to load configuration.", e);
							this.skipRemainigCommands();
						}
					}
					else{
						this.outputStream.println("Illegal argument for config command.");
						this.skipRemainigCommands();
					}
					
					this.info("Finished configuration.");
				}
				else{
					this.info("Unknown command: {}", command);
					
					this.skipRemainigCommands();
					this.outputStream.println("Sorry, I do not know that command! Please enter another one.");
				}
			}
			
			this.info("Finished start.");
			
			while(running){
				try{
					try{
						command = this.nextCommand().trim();
						} catch (EndOfInputException e) {
						running = false;
						this.info("End of input.",e);
						break;
					}
					this.debug("Executing command: " + command);
					if(command.equals("")){
						continue;
					}
					else if(command.startsWith("config")){
						this.info("Starting configuration.");
						
						Matcher matcher = configCommandPattern.matcher(command);
						
						if(matcher.matches()){
							String configFile = matcher.group(1);
							this.info("Loading configuration.");						
							try{
								if(Configuration.existsFile(configFile)){
									this.info("Configuration exists. Loading configuration.");
									
									try {
										this.config(Configuration.readFile(configFile));
										this.outputStream.println("Loaded configuration");
										this.info("Loaded configuration.");
									} catch (IOException e) {
										this.outputStream.println("Unable to load configuration.");
										this.warn("Unable to load configuration.", e);
										this.skipRemainigCommands();
									}
									
								}
								else{
									this.outputStream.println("Configuration file does not exist: " + new File(configFile).getAbsolutePath());
									this.info("Configuration file does not exist: {}", new File(configFile).getAbsolutePath());
									this.skipRemainigCommands();
								}
							} catch (Exception e) {
								this.outputStream.println("Unable to load configuration.");
								this.warn("Unable to load configuration.", e);
								this.skipRemainigCommands();
							}
						}
						else{
							this.outputStream.println("Illegal argument for config command.");
							this.skipRemainigCommands();
						}
						
						this.info("Finished configuration.");
					}
					else if(command.startsWith("compile")){
						this.info("Starting compilation.");
						
						Matcher matcher = compileCommandPattern.matcher(command);
						
						if(matcher.matches()){
							String aliasesList = matcher.group(1);
							String namespace = matcher.group(2);
							String simple_name = matcher.group(3);
							String argumentList = matcher.group(4);
							
							ArrayList<String> aliases = new ArrayList<String>();
							
							Matcher aliasesMatcher = Pattern.compile(compilerAliasRegEx).matcher(aliasesList);
							
							while(aliasesMatcher.find()){
								aliases.add(aliasesMatcher.group());
							}
							
							
							ArrayList<String> arguments = new ArrayList<String>();
							if(argumentList.startsWith(",")){
								arguments.add("");
								argumentList = argumentList.substring(1);
							}
							
							Matcher argumentMatcher = Pattern.compile(fileNameRegEx).matcher(argumentList);
							
							while(argumentMatcher.find()){
								arguments.add(argumentMatcher.group());
							}
							
							String destination = (arguments.size() > 1 ? arguments.get(0) : "");
							
							assert arguments.size() >= 1;							
							String[] compilerArguments;
							if(arguments.size() == 1){
								compilerArguments = new String[]{arguments.get(0)};
							}
							else{
								compilerArguments = new String[arguments.size() - 1];
								System.arraycopy(arguments.toArray(new String[arguments.size()]), 1, compilerArguments, 0, compilerArguments.length);
							}
							
							DTSA_Program dtsa_program = null;
							try {
								dtsa_program = this.eventMill.compile(aliases.toArray(new String[aliases.size()]), namespace, simple_name, compilerArguments);
															
								this.info("Finished compilation.");
								
								this.outputStream.println("Compilation succeeded.");
							} catch (DuraCompilationException e) {
								this.info("Compilation did not succeed", e);
								
								this.outputStream.println("Compilation did not succeed.");
								this.outputStream.println(e.getClass().getSimpleName() + ": " + e.getMessage());
								
								this.skipRemainigCommands();
							} catch (Exception e) {
								this.info("Compilation did not succeed", e);
								
								this.outputStream.println("Compilation did not succeed.");
								this.outputStream.println(e.getClass().getSimpleName() + ": " + e.getMessage());
								e.printStackTrace(this.outputStream);
								this.skipRemainigCommands();
							}
							if(dtsa_program != null){
								if(destination == null){
									try{
										this.eventMill.addSpec(dtsa_program);
										
										this.info("Loaded program specification.");
										
										this.outputStream.println("Loaded program specification.");
									}
									catch (IllegalArgumentException e){							
										this.info("",e);
										
										e.printStackTrace(this.outputStream);
										this.outputStream.println();
										this.skipRemainigCommands();
										
									}
									catch (Exception e) {
										this.error("",e);
										
										e.printStackTrace(this.outputStream);
										this.outputStream.println();
										this.skipRemainigCommands();
									}
								}
								else{
									this.eventMill.save(dtsa_program, destination);
								}
							}
													
						}
						else{
							this.outputStream.println("Illegal argument for compile command.");
							this.skipRemainigCommands();
						}
					}
					else if(command.startsWith("load spec")){
						this.info("Loading program specification.");
						
						Matcher matcher = loadSpecCommandPattern.matcher(command);
		
						if(matcher.matches()){
							String path = matcher.group(1);
							
							DTSA_Program dtsa_program = null;
							try {
								dtsa_program = this.eventMill.load(path);
								
							} catch (Exception e) {
								this.info("File could not be read.", e);
								
								this.outputStream.println("File could not be read: " + new File(path).getAbsolutePath());
								this.outputStream.println(e.getLocalizedMessage());
								this.skipRemainigCommands();
							}
							if(dtsa_program != null){
								
								try{
									this.eventMill.addSpec(dtsa_program);
									this.info("Loaded program specification.");
									
									this.outputStream.println("Loaded program specification.");
								}
								catch (IllegalArgumentException e){							
									this.info("",e);
									
									e.printStackTrace(this.outputStream);
									this.outputStream.println();
									this.outputStream.println("Did not load program specification.");
									this.skipRemainigCommands();
								}
								catch (Exception e) {
									this.error("",e);
									
									e.printStackTrace(this.outputStream);
									this.outputStream.println();
									this.outputStream.println("Did not load program specification.");
									this.skipRemainigCommands();
								}
								
							}
													
						}
						else{
							this.outputStream.println("Illegal argument for load spec command.");
							this.skipRemainigCommands();
						}
					}
					else if(command.matches(listSpecsCommandRegEx)){
	
						this.outputStream.println("The following program specifications are available:");
						
						for(String programSpecName: this.eventMill.availableProgramSpecifications()){
							this.outputStream.println(programSpecName);
						}
						this.outputStream.println();
						
					}
					else if(command.matches(clearSpecsCommandRegEx)){
						this.info("Clearing program specifications.");
						
						this.eventMill.clearSpecs();
						
						this.outputStream.println("Cleared program specifications.");
						
						this.info("Cleared program specifications.");
						this.outputStream.println("Cleared program specifications.");
					}
					else if(command.startsWith("remove spec")){
						this.info("Removing program specification.");
						
						Matcher matcher = removeSpecCommandPattern.matcher(command);
		
						if(matcher.matches()){
							String full_name = matcher.group(1);
							
							this.eventMill.removeSpec(full_name);
								
							this.info("Removed program specification \"{}\".", full_name);
	
							this.outputStream.println("Removed program specification \"" + full_name + "\".");
						}
						else{
							this.outputStream.println("Illegal argument for remove spec command.");
							this.skipRemainigCommands();
						}
					}
					else if(command.startsWith("init")){	
						this.info("Initializing program specification.");
						
						Matcher matcher = initCommandPattern.matcher(command);
		
						if(matcher.matches()){
							String[] full_instance_name = TSA_SyntaxPatterns.programInstFullName(matcher.group(1));
							String full_name = TSA_SyntaxPatterns.programFullName(full_instance_name[0], full_instance_name[1]);
							String instance = full_instance_name[2];
							
							//FIXME arbitrary parameters like for compile
							String meta_schema = matcher.group(5);
							String input_schema = matcher.group(6);
							String working_schema = matcher.group(7);
							String output_schema = matcher.group(8);
							String log_schema = matcher.group(9);
							
							String[] parameters = new String[]{meta_schema, input_schema, working_schema, output_schema, log_schema};
							
							if(this.eventMill.hasSpec(full_name)){
								
								try{
									this.eventMill.instanciate(full_name, instance, parameters);
									
									this.outputStream.println("Initialized program specification.");
									this.info("Initialized program specification.");
								}
								catch (IllegalArgumentException e){							
									this.info("",e);
									
									e.printStackTrace(this.outputStream);
									this.outputStream.println();
									this.skipRemainigCommands();
								}
								catch (Exception e) {
									this.error("",e);
									
									e.printStackTrace(this.outputStream);
									this.outputStream.println();
									this.skipRemainigCommands();
								}
							}
							else{
								this.info("There is no program specification with name \"{}\" available.", full_name);
								
								this.outputStream.println("There is no program specification with name \"" + full_name + "\" available.");
								this.skipRemainigCommands();
							}
								
						}
						else{
							this.info("Illegal argument for init command.");
							
							this.outputStream.println("Illegal argument for init command.");
							this.skipRemainigCommands();
						}
					}
					else if(command.matches(listProgsCommandRegEx)){
						this.outputStream.println("The following initialized programs are available:");
						
						for(String programName: this.eventMill.availablePrograms()){
							this.outputStream.println(programName);
						}
						this.outputStream.println();
						
					}
					else if(command.matches(clearProgsCommandRegEx)){
	
						this.info("Clearing program instances.");
						
						this.eventMill.clearProgs();
						
						this.outputStream.println("Cleared program instances");
						
						this.info("Cleared program instances.");
						
						this.outputStream.println("Cleared program instances.");
					}
					else if(command.startsWith("remove prog")){
						
						Matcher matcher = removeInstCommandPattern.matcher(command);
		
						if(matcher.matches()){
							String full_name = matcher.group(1);
							
							this.info("Removing program instance: {}", full_name);
							
							this.eventMill.removeProg(full_name);
								
							this.info("Removed program instance: {}", full_name);
	
							this.outputStream.println("Removed program instance: " + full_name);
						}
						else{
							this.info("Illegal argument for remove inst command.");
							
							this.outputStream.println("Illegal argument for remove inst command.");
							this.skipRemainigCommands();
						}
					}
					else if(command.matches(clearCommandRegEx)){
	
						this.info("Clearing program instances and program specifications.");
						
						this.eventMill.clear();
		
						this.info("Cleared program instances and program specifications.");
		
						this.outputStream.println("Cleared program instances and program specifications.");
					}
					else if(command.startsWith("run")){
	
						Matcher runCommandMatcher = Pattern.compile(runCommandRegEx).matcher(command);
						if(runCommandMatcher.matches()){
							String program_full_name = runCommandMatcher.group(1);
							
							if(this.eventMill.availablePrograms().contains(program_full_name)){
							
								ExecutableProgram<?,?,?> program = null;
								try{
									program = this.eventMill.getProg(program_full_name);
								}
								catch (Exception e) {
									this.error("",e);
									
									e.printStackTrace(this.outputStream);
									this.outputStream.println();
									this.skipRemainigCommands();
								}
								if(program != null){
								
									ControlThread controlThread = new ControlThread(this.variables, this.eventMill.configuration().scheduler, program, this.bufferedCommands, this.inputStream, this.outputStream);
									if(controlThread.skipedCommands()){
										this.skipBufferedCommands();
									}
					
									this.outputStream.print("Loaded: ");
									this.outputStream.println(program_full_name);
									
									this.outputStream.print("Switching to: ");
									this.outputStream.println(program_full_name);
									controlThread.start();
									while(controlThread.isAlive()){
										try{
											controlThread.join();
										}
										catch (InterruptedException e) {
				
										}
									}
									if(controlThread.skipedCommands()){
										this.skipBufferedCommands();
									}
									
									this.outputStream.print("Returned from: ");
									this.outputStream.println(program_full_name);
									
								}
							}
							else{
								this.info("Program \"{}\" is not available.", program_full_name);
								this.outputStream.println("Program \"" + program_full_name + "\" is not available.");
								this.skipRemainigCommands();
							}
						}
						else{
							this.info("Illegal argument for run command.");
							
							this.outputStream.println("Illegal argument for run command.");
							this.skipRemainigCommands();
						}
					}
					else if(command.startsWith("pause")){
						Matcher matcher = pauseCommandPattern.matcher(command);
						if(matcher.matches()){
							if(matcher.group(1) == null){
								this.outputStream.println("Press [Enter] to continue.");
								this.pause();
							}
							else{
								long milies = Long.parseLong(matcher.group(1));
								this.outputStream.println("Pausing for " + milies + " millies.");
								this.pause(milies);
							}
						}
						else{
							this.outputStream.println("Illegal argument for pause command.");
							this.skipRemainigCommands();
						}
					}
					else if(command.startsWith("set")){
						Matcher matcher = variableDefinitionPattern.matcher(command);
						
						if(matcher.matches()){
							String var = matcher.group(1);
							String value = matcher.group(2);
							this.setVariable(var, value);
						}
						else{
							this.outputStream.println("Illegal argument for set command.");
							this.skipRemainigCommands();
						}
					}
					else if(command.startsWith("inject")){
						Matcher matcher = injectCommandPattern.matcher(command);
						if(matcher.matches()){
							String file = matcher.group(1);
							try {
								this.inject(file);
							} catch (FileNotFoundException e) {
								this.outputStream.println("Unable to load file \"" + new File(file).getAbsolutePath() + "\".");
								this.info("Unable to load file.", e);
								this.skipRemainigCommands();
							}
						}
						else{
							this.outputStream.println("Illegal argument for pause command.");
							this.skipRemainigCommands();
						}
					}
					else if("quit".equalsIgnoreCase(command)){
						running = false;
						
						this.debug("Stopping...");
					}
					else if(command.matches(uninstallCommandRegEx)){
	
						running = false;
						this.info("Uninstalling.");
						
						this.eventMill.uninstall();
		
						this.info("Uninstalled.");
		
						this.outputStream.println("Uninstalled.");
						
						this.debug("Stopping...");				
					}
					else{
						this.info("Unknown command: {}", command);
						
						this.outputStream.println("Sorry, I do not know that command! Please enter another one.");
						this.skipRemainigCommands();
					}
				}
				catch (Exception e) {
					this.outputStream.println(e.toString());
					this.error("Exception", e);
					this.skipRemainigCommands();
				}
			}
			
			this.outputStream.println("Bye!");
			
			this.info("Stopped.");
		
		}
		catch(Throwable t){
			this.error("",t);
			t.printStackTrace(this.outputStream);
			exitCode = 1;
		}
		finally{
			if(this.eventMill != null){
				this.eventMill.close();
			}
			try {
				Thread.sleep(500);
			} catch (InterruptedException e) {

			}
			System.exit(exitCode);
		}
	}
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(MainThread.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
