package de.lmu.ifi.pms.cep.event_mill.execution.engine;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;

import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgram;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgramState;
import de.lmu.ifi.pms.cep.event_mill.execution.scheduler.BasicQueryThreadFactory;
import de.lmu.ifi.pms.cep.event_mill.execution.scheduler.QueryThread;
import de.lmu.ifi.pms.cep.event_mill.execution.scheduler.QueryThreadFactory;
import de.lmu.ifi.pms.cep.event_mill.execution.scheduler.SchedulerFactory;
import de.lmu.ifi.pms.cep.event_mill.execution.scheduler.SchedulerThread;

import de.lmu.ifi.pms.java.util.stream.io.AbstractCloseable;



public class Engine extends AbstractCloseable{

	protected EngineState state;
	
	protected SchedulerFactory schedulerFactory;
	protected QueryThreadFactory queryThreadFactory;
	
	protected SchedulerThread schedulerThread;
	protected QueryThread queryThread;
		
	public final ExecutableProgram<?,?,?> executableProgram;
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(Engine.class);	
	public final Marker marker;
	
	public Engine(final ExecutableProgram<?,?,?> executableProgram)
	{
		this.executableProgram = executableProgram;
		
		this.marker = this.executableProgram.marker;
		
		this.state = EngineState.CREATED;
		this.info("{}", this.state);
	}
	
	public Engine(final ExecutableProgram<?,?,?> executableProgram, final SchedulerFactory schedulerFactory){
		this(executableProgram);
		
		this.init(schedulerFactory);
	}
	
	public Engine(final ExecutableProgram<?,?,?> executableProgram, final SchedulerFactory schedulerFactory, final QueryThreadFactory queryThreadFactory)
	{
		this(executableProgram);
		
		this.init(schedulerFactory, queryThreadFactory);
	}
	
	public synchronized void init(final SchedulerFactory schedulerFactory){
		//TODO generic query thread factory
		this.init(schedulerFactory, new BasicQueryThreadFactory());
	}
	
	public synchronized void init(final SchedulerFactory schedulerFactory, final QueryThreadFactory queryThreadFactory)
	{
		if(this.state == EngineState.CREATED)
		{
			try{
				this.schedulerFactory = schedulerFactory;
				this.queryThreadFactory = queryThreadFactory;
				
				this.executableProgram.lock();
				
				if(!this.executableProgram.isInitialized()){
					this.executableProgram.init();
				}
				
				this.state = EngineState.INITIALIZED;
				this.notifyAll();
				this.info("{}", this.state);
			} catch (RuntimeException e) {
				this.executableProgram.unlock();
				throw e;
			} catch (Error e) {
				this.executableProgram.unlock();
				throw e;
			}
		}
		else{
			throw new IllegalStateException();//TODO better exception
		}
	}
	
	public synchronized boolean isInitialized(){
		return this.state != EngineState.CREATED;
	}
	
	public synchronized void start()
	{
		if(this.state == EngineState.INITIALIZED || this.state == EngineState.STOPPED)
		{
			this.executableProgram.run();
			this.schedulerThread = this.schedulerFactory.make(this);
			this.queryThread = this.queryThreadFactory.make(this.schedulerThread);			
			this.schedulerThread.start();
			this.queryThread.start();
			
			this.state = EngineState.RUNNING;
			this.notifyAll();
			this.info("{}", this.state);
		}
		else{
			throw new IllegalStateException();//TODO better exception
		}
	}
	
	public synchronized boolean isRunning(){
		return this.state == EngineState.RUNNING;
	}
	
	public synchronized boolean isStopping(){
		return this.state == EngineState.STOPPING;
	}
	
	public void stop()
	{
		Thread schedulerThread;
		Thread queryThread;
		synchronized(this){
			if(this.state == EngineState.RUNNING)
			{
				schedulerThread = this.schedulerThread;
				queryThread = this.queryThread;
				
				this.state = EngineState.STOPPING;
				this.notifyAll();
				this.info("{}", this.state);
			}
			else{
				throw new IllegalStateException();//TODO better exception
			}
		}
		while(schedulerThread.isAlive()){
			try
			{
				schedulerThread.join();
			}
			catch (InterruptedException e)
			{
				
			}
		}
		while(queryThread.isAlive()){
			try
			{
				queryThread.join();
			}
			catch (InterruptedException e)
			{
				
			}
		}
		synchronized(this){
			
			this.schedulerThread = null;
			this.queryThread = null;
			
			if(this.state != EngineState.CLOSING){
				this.executableProgram.stop();
				this.state = EngineState.STOPPED;
				this.notifyAll();
				this.info("{}", this.state);
			}
		}
	}
	
	public synchronized boolean isStopped(){
		return this.state == EngineState.STOPPED;
	}
	
	public synchronized void reset(){
		if(this.state == EngineState.INITIALIZED || this.state == EngineState.STOPPED)
		{ 
			this.schedulerFactory = null;
			this.queryThreadFactory = null;
			
			this.executableProgram.unlock();
			
			this.state = EngineState.CREATED;
			this.notifyAll();
			this.info("{}", this.state);
		}
		else{
			throw new IllegalStateException();//TODO better exception
		}
	}
	
	public void close(final Throwable e){
		this.error("Exceptional shutdown of engine:", e);
		this.close();
		/*if(e instanceof RuntimeException){
			throw (RuntimeException) e;
		}
		if(e instanceof Error){
			throw (Error) e;
		}
		throw new RuntimeException(e);*/
	}
	
	@Override
	public void close() {
		EngineState lastState;
		Thread schedulerThread;
		Thread queryThread;
		
		synchronized(this){
			lastState = this.state;
			schedulerThread = this.schedulerThread;
			queryThread = this.queryThread;
			if(lastState != EngineState.CLOSING & lastState != EngineState.CLOSED){
				this.state = EngineState.CLOSING; //isRunning() will fail now -> Scheduler terminates
				this.notifyAll();
				this.info("{}", this.state);
			}
		}
		
		if(lastState == EngineState.RUNNING || lastState == EngineState.STOPPING){
			while(schedulerThread.isAlive()){
				try
				{
					schedulerThread.join();
				}
				catch (InterruptedException e)
				{
					
				}
			}
			while(queryThread.isAlive()){
				try
				{
					queryThread.join();
				}
				catch (InterruptedException e)
				{
					
				}
			}
		}

		synchronized(this){
			assert this.state == EngineState.CLOSING | this.state == EngineState.CLOSED;
			if(this.state != EngineState.CLOSED){
				this.schedulerThread = null;
				this.queryThread = null;
				
				this.schedulerFactory = null;
				this.queryThreadFactory = null;
			
				if(lastState != EngineState.CREATED ){
					if(lastState != EngineState.INITIALIZED & lastState != EngineState.STOPPED){
						synchronized (this.executableProgram) {
							if(this.executableProgram.State() == ExecutableProgramState.RUNNING){
								this.executableProgram.stop();
							}
						}
					}
					this.executableProgram.unlock();
				}
				
				this.state = EngineState.CLOSED;
				this.notifyAll();
				this.info("{}", this.state);
			}
		}
	}
	
	public synchronized EngineState State()
	{
		return this.state;
	}
	
	//logging
	
	protected Logger logger(){
		return logger;
	}

	protected final void trace(final String msg) {
		this.logger().trace(this.marker, msg);
	}

	protected final void trace(final String format, final Object arg) {
		this.logger().trace(this.marker, format, arg);
	}

	protected final void trace(final String format, final Object arg1, final Object arg2) {
		this.logger().trace(this.marker, format, arg1, arg2);
	}

	protected final void trace(final String format, final Object[] argArray) {
		this.logger().trace(this.marker, format, argArray);
	}

	protected final void trace(final String msg, final Throwable t) {
		this.logger().trace(this.marker, msg, t);
	}

	protected final void debug(final String msg) {
		this.logger().debug(this.marker, msg);
	}

	protected final void debug(final String format, final Object arg) {
		this.logger().debug(this.marker, format, arg);
	}

	protected final void debug(final String format, final Object arg1, final Object arg2) {
		this.logger().debug(this.marker, format, arg1, arg2);
	}

	protected final void debug(final String format, final Object[] argArray) {
		this.logger().debug(this.marker, format, argArray);
	}

	protected final void debug(final String msg, final Throwable t) {
		this.logger().debug(this.marker, msg, t);
	}

	protected final void info(final String msg) {
		this.logger().info(this.marker, msg);
	}

	protected final void info(final String format, final Object arg) {
		this.logger().info(this.marker, format, arg);
	}

	protected final void info(final String format, final Object arg1, final Object arg2) {
		this.logger().info(this.marker, format, arg1, arg2);
	}

	protected final void info(final String format, final Object[] argArray) {
		this.logger().info(this.marker, format, argArray);
	}

	protected final void info(final String msg, final Throwable t) {
		this.logger().info(this.marker, msg, t);
	}

	protected final void warn(final String msg) {
		this.logger().warn(this.marker, msg);
	}

	protected final void warn(final String format, final Object arg) {
		this.logger().warn(this.marker, format, arg);
	}

	protected final void warn(final String format, final Object arg1, final Object arg2) {
		this.logger().warn(this.marker, format, arg1, arg2);
	}

	protected final void warn(final String format, final Object[] argArray) {
		this.logger().warn(this.marker, format, argArray);
	}

	protected final void warn(final String msg, final Throwable t) {
		this.logger().warn(this.marker, msg, t);
	}

	protected final void error(final String msg) {
		this.logger().error(this.marker, msg);
	}

	protected final void error(final String format, final Object arg) {
		this.logger().error(this.marker, format, arg);
	}

	protected final void error(final String format, final Object arg1, final Object arg2) {
		this.logger().error(this.marker, format, arg1, arg2);
	}

	protected final void error(final String format, final Object[] argArray) {
		this.logger().error(this.marker, format, argArray);
	}

	protected final void error(final String msg, final Throwable t) {
		this.logger().error(this.marker, msg, t);
	}
	
}
