package de.lmu.ifi.pms.cep.event_mill.execution.program;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.compilation.CompilerFactory;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;


public abstract class StreamBufferStore<StoreType extends StreamBufferStore<StoreType, SpecType>, SpecType> {
	
	public final Class<StoreType> type;
	public final Class<SpecType> programSpecType;
	
	public StreamBufferStore(final Class<StoreType> type, final Class<SpecType> programSpecType) {
		super();
		this.type = type;
		this.programSpecType = programSpecType;
	}
	
	public abstract CompilerFactory<DTSA_Program, SpecType> DTSASpecCompiler();

	public abstract void initialize();
	
	public abstract boolean isInizialized();
	
	public abstract void remove();
	
	public abstract FixedArrayNavigableSet<String> availableProgramSpecifications();
	
	protected abstract SpecType getProgramSpecification(final String namespace, final String simple_name);
	
	protected SpecType getProgramSpecification(final String full_name){
		String[] decoded = TSA_SyntaxPatterns.programFullName(full_name);
		return this.getProgramSpecification(decoded[0], decoded[1]);
	}
	
	public void add(final DTSA_Program dtsa_program){
		this.add(this.DTSASpecCompiler().make().compile(dtsa_program));
	}
	
	protected abstract void add(final SpecType program_spec);
	
	public boolean existsProgSpec(final String full_name){
		String[] name = TSA_SyntaxPatterns.programFullName(full_name);
		return this.existsProgSpec(name[0], name[1]);
	}
	
	public abstract boolean existsProgSpec(final String namespace, final String simple_name);
	
	public void removeProgramSpec(final String full_name){
		String[] name = TSA_SyntaxPatterns.programFullName(full_name);
		this.removeProgramSpec(name[0], name[1]);
	}
	
	public abstract void removeProgramSpec(final String namespace, final String simple_name);
	
	public void clearProgramSpecs(){
		//TODO easier implementation
		for(String full_name: this.availableProgramSpecifications()){
			this.removeProgramSpec(full_name);
		}
	}
	
	public void instanciate(final String namespace, final String simple_name, final String instance, final String... parameters){
		SpecType program_spec = this.getProgramSpecification(namespace, simple_name);
		this.instanciate(program_spec, instance, parameters);
	}
	
	protected abstract void instanciate(final SpecType program_spec, final String instance, final String... parameters);
	
	public boolean existsProgInst(final String full_name){
		String[] name = TSA_SyntaxPatterns.programInstFullName(full_name);
		return this.existsProgInst(name[0], name[1], name[2]);
	}
	
	public void removeProgramInst(final String full_name){	
		String[] name = TSA_SyntaxPatterns.programInstFullName(full_name);
		this.removeProgramInst(name[0], name[1], name[2]);
	}
	
	public abstract void removeProgramInst(final String namespace, final String simple_name, final String instance);
	
	public abstract FixedArrayNavigableSet<String> availablePrograms();
	
	public void clearProgramInsts(){
		for(String program_full_name: this.availablePrograms()){
			this.removeProgramInst(program_full_name);
		}
	}
	
	public abstract boolean existsProgInst(final String namespace, final String simple_name, final String instance);
	
	public ExecutableProgram<StoreType, ?, ?> getProgram(final String full_name){
		String[] name = TSA_SyntaxPatterns.programInstFullName(full_name);
		return this.getProgram(name[0], name[1], name[2]);
	}
	
	public abstract ExecutableProgram<StoreType, ?, ?> getProgram(final String namespace, final String simple_name, final String instance);

	public abstract void close();
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(StreamBufferStore.class);
	
	//logging
	
	protected Logger logger(){
		return logger;
	}

	protected final void trace(final String msg) {
		this.logger().trace(msg);
	}

	protected final void trace(final String format, final Object arg) {
		this.logger().trace(format, arg);
	}

	protected final void trace(final String format, final Object arg1, final Object arg2) {
		this.logger().trace(format, arg1, arg2);
	}

	protected final void trace(final String format, final Object[] argArray) {
		this.logger().trace(format, argArray);
	}

	protected final void trace(final String msg, final Throwable t) {
		this.logger().trace(msg, t);
	}

	protected final void debug(final String msg) {
		this.logger().debug(msg);
	}

	protected final void debug(final String format, final Object arg) {
		this.logger().debug(format, arg);
	}

	protected final void debug(final String format, final Object arg1, final Object arg2) {
		this.logger().debug(format, arg1, arg2);
	}

	protected final void debug(final String format, final Object[] argArray) {
		this.logger().debug(format, argArray);
	}

	protected final void debug(final String msg, final Throwable t) {
		this.logger().debug(msg, t);
	}

	protected final void info(final String msg) {
		this.logger().info(msg);
	}

	protected final void info(final String format, final Object arg) {
		this.logger().info(format, arg);
	}

	protected final void info(final String format, final Object arg1, final Object arg2) {
		this.logger().info(format, arg1, arg2);
	}

	protected final void info(final String format, final Object[] argArray) {
		this.logger().info(format, argArray);
	}

	protected final void info(final String msg, final Throwable t) {
		this.logger().info(msg, t);
	}

	protected final void warn(final String msg) {
		this.logger().warn(msg);
	}

	protected final void warn(final String format, final Object arg) {
		this.logger().warn(format, arg);
	}

	protected final void warn(final String format, final Object arg1, final Object arg2) {
		this.logger().warn(format, arg1, arg2);
	}

	protected final void warn(final String format, final Object[] argArray) {
		this.logger().warn(format, argArray);
	}

	protected final void warn(final String msg, final Throwable t) {
		this.logger().warn(msg, t);
	}

	protected final void error(final String msg) {
		this.logger().error(msg);
	}

	protected final void error(final String format, final Object arg) {
		this.logger().error(format, arg);
	}

	protected final void error(final String format, final Object arg1, final Object arg2) {
		this.logger().error(format, arg1, arg2);
	}

	protected final void error(final String format, final Object[] argArray) {
		this.logger().error(format, argArray);
	}

	protected final void error(final String msg, final Throwable t) {
		this.logger().error(msg, t);
	}
	
}
