package de.lmu.ifi.pms.cep.event_mill.sql.io;

import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.NoSuchElementException;

import org.simpleframework.xml.core.Persister;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.configuration.Configuration;
import de.lmu.ifi.pms.cep.event_mill.configuration.DEFAULTS;
import de.lmu.ifi.pms.cep.event_mill.configuration.SQL_ConnectionConfiguration;
import de.lmu.ifi.pms.cep.event_mill.configuration.SQL_DatabaseDriver;
import de.lmu.ifi.pms.cep.event_mill.io.EngineIO;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SinkStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SourceStream;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_EXCEPTIONS;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_StreamBufferStore;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_StreamBufferStoreFactory;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Translation;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;



public class SQL_EngineIO extends EngineIO<SQL_StreamBufferStore>{
	
	final SQL_DatabaseDriver driver;
	public final SQL_Translation sql_translation;
	final SQL_ConnectionConfiguration connectionConfiguration;
	Connection connection;
	Statement statement;
	boolean usePreparedStatements;
	
	final String engine_meta_schema;
	
	final String program_meta_schema;
	final String program_input_schema;
	final String program_output_schema;
	final String program_log_schema;
	
	int maxNumberOfInputStreamBounds;
	int maxNumberOfOutputStreamBounds;
	int maxNumberOfReferencedClocks;
	
	final String getBufferSimpleEntrySQL;
	PreparedStatement getBufferSimpleEntry;
	final String getOutputQueryEntrySQL;
	PreparedStatement getOutputQueryEntry;
	final String getBufferSimpleEntriesOfTypeSQL;
	PreparedStatement getBufferSimpleEntriesOfType;
	final String getBufferSimpleEntriesOfTypeAndModuleSQL;
	PreparedStatement getBufferSimpleEntriesOfTypeAndModule;
	final String getAttributeEntrySQL;
	PreparedStatement getAttributeEntry;
	final String getAttributeEntriesSQL;
	PreparedStatement getAttributeEntries;
	
	public SQL_EngineIO(final String streamBufferStoreIdentifier, final String program_full_name) throws IOException{
		this(streamBufferStoreIdentifier, DEFAULTS.defaultConfigResourcePath, program_full_name);
	}
	
	public SQL_EngineIO(final String streamBufferStoreIdentifier, final String configurationFilePath, final String program_full_name) throws IOException{
		this(streamBufferStoreIdentifier, (SQL_StreamBufferStoreFactory)Configuration.readFile(configurationFilePath).streamBufferStore, program_full_name);
	}
	
	public SQL_EngineIO(final String streamBufferStoreIdentifier, final SQL_StreamBufferStoreFactory configuration, final String program_full_name) {
		this(streamBufferStoreIdentifier, configuration.driver, configuration.connection, configuration.meta_schema, program_full_name);
	}
	
	public SQL_EngineIO(final String streamBufferStoreIdentifier, final SQL_DatabaseDriver driver, final SQL_ConnectionConfiguration connection, final String engine_meta_schema, final String program_full_name) {
		super(SQL_StreamBufferStore.class, streamBufferStoreIdentifier, program_full_name);
		this.driver = driver;
		this.sql_translation = driver.translation();
		this.connectionConfiguration = connection;
		this.usePreparedStatements = connection.usePreparedStatements;
		this.engine_meta_schema = engine_meta_schema;
		this.connect();
	
		String getProgramInstSQL = this.sql_translation.sqlGetProgramInst(this.engine_meta_schema, program_full_name);
		
		String[] schemas;
		int maxNumberOfInputStreamBounds;
		int maxNumberOfOutputStreamBounds;
		int maxNumberOfReferencedClocks;
		while(true){
			try{
				ResultSet program_instance = this.statement.executeQuery(getProgramInstSQL);
				if(program_instance.next()){						
					schemas = this.sql_translation.sqlProgramInstReadSchemas(program_instance);
					
					maxNumberOfInputStreamBounds = this.sql_translation.sqlProgramInstReadQueryMaxNumberOfInputStreamBounds(program_instance);
					maxNumberOfOutputStreamBounds = this.sql_translation.sqlProgramInstReadQueryMaxNumberOfOutputStreamBounds(program_instance);
					maxNumberOfReferencedClocks = this.sql_translation.sqlProgramInstReadQueryMaxNumberOfReferencedClocks(program_instance);
					try{program_instance.close();} catch(SQLException e){};
				}
				else{
					try{program_instance.close();} catch(SQLException e){};
					throw new RuntimeException("Program instance does not exist: " + this.identifier);
				}
				
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
					continue;
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
		
		this.maxNumberOfInputStreamBounds = maxNumberOfInputStreamBounds;
		this.maxNumberOfOutputStreamBounds = maxNumberOfOutputStreamBounds;
		this.maxNumberOfReferencedClocks = maxNumberOfReferencedClocks;
		
		this.program_meta_schema = schemas[0];
		this.program_input_schema = schemas[1];
		this.program_output_schema = schemas[3];
		this.program_log_schema = schemas[4];
		
		this.getBufferSimpleEntrySQL = this.sql_translation.sqlGetBufferSimpleMetaEntry(this.program_meta_schema);
		this.getOutputQueryEntrySQL = this.sql_translation.sqlGetOutputQueryMetaEntry(this.program_meta_schema, this.maxNumberOfOutputStreamBounds, this.maxNumberOfInputStreamBounds, this.maxNumberOfReferencedClocks);
		this.getBufferSimpleEntriesOfTypeSQL = this.sql_translation.sqlGetBufferMetaSimpleEntriesOfType(this.program_meta_schema);
		this.getBufferSimpleEntriesOfTypeAndModuleSQL = this.sql_translation.sqlGetBufferMetaSimpleEntriesOfTypeAndModule(this.program_meta_schema);
		this.getAttributeEntrySQL = this.sql_translation.sqlGetAttributeMetaEntry(this.program_meta_schema);
		this.getAttributeEntriesSQL = this.sql_translation.sqlGetAttributeMetaEntries(this.program_meta_schema);
		
		this.initStatements();
	}
	
	protected void connect(){
		if(this.connection != null ){
			assert this.statement != null;
			try {this.statement.close();} catch (SQLException e) {}
			try {this.connection.close();} catch (SQLException e) {}
			this.statement = null;
			this.connection = null;
		}
		
		this.info("Connecting to database...");
		
		try{
			this.connection = this.driver.connect(this.connectionConfiguration);
		} catch(SQLException e1){
			this.error("Error on connecting to database!", e1);
			throw new RuntimeException(e1);
		}
		this.info("Connected to database.");

		try{
			this.statement = this.connection.createStatement();
		} catch(SQLException e){
			this.error("Error on creating statement!", e);
			throw new RuntimeException(e);
		}
	}
	
	public synchronized void reconnect(final Connection currentConnection){
		if(this.connection == currentConnection){
			this.connect();
			this.initStatements();
		}
	}
	
	protected synchronized void initStatements(){
		while(true){
			try{				
				this.getBufferSimpleEntry = this.sql_translation.prepareStatement(this.connection, this.getBufferSimpleEntrySQL, this.usePreparedStatements);
				this.getOutputQueryEntry = this.sql_translation.prepareStatement(this.connection, this.getOutputQueryEntrySQL, this.usePreparedStatements);
				this.getBufferSimpleEntriesOfType = this.sql_translation.prepareStatement(this.connection, this.getBufferSimpleEntriesOfTypeSQL, this.usePreparedStatements);
				this.getBufferSimpleEntriesOfTypeAndModule = this.sql_translation.prepareStatement(this.connection, this.getBufferSimpleEntriesOfTypeAndModuleSQL, this.usePreparedStatements);
				this.getAttributeEntry = this.sql_translation.prepareStatement(this.connection, this.getAttributeEntrySQL, this.usePreparedStatements);
				this.getAttributeEntries = this.sql_translation.prepareStatement(this.connection, this.getAttributeEntriesSQL, this.usePreparedStatements);
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect(this.connection);
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	public synchronized SQL_DatabaseDriver driver(){
		return this.driver;
	}
	
	public synchronized SQL_ConnectionConfiguration connection(){
		return this.connectionConfiguration;
	}
	
	public synchronized String metaSchema(){
		return this.program_meta_schema;
	}
	
	public synchronized String inputSchema(){
		return this.program_input_schema;
	}
	
	public synchronized String outputSchema(){
		return this.program_output_schema;
	}
	
	public synchronized String logSchema(){
		return this.program_log_schema;
	}
	
	public static String programInstFullName(final String namespace, final String simple_name, final String instance){
		return TSA_SyntaxPatterns.programInstFullName(namespace, simple_name, instance);
	}
	
	public static void checkBufferType(final DTSA_BufferType buffer_type){
		if(buffer_type == DTSA_BufferType.WORKING){
			throw new IllegalArgumentException("Information about internal buffers should not be accessed");
		}
	}
	
	@Override
	public synchronized FixedArrayNavigableSet<String> bufferNames(final DTSA_BufferType buffer_type){
		checkBufferType(buffer_type);
		ResultSet buffers;
		while(true){
			try {
				this.sql_translation.sqlSetGetBufferMetaEntriesOfType(this.getBufferSimpleEntriesOfType, buffer_type);
				buffers = this.getBufferSimpleEntriesOfType.executeQuery();
				FixedArrayNavigableSet<String> bufferNames = this.sql_translation.sqlReadBufferMetaEntriesNames(buffers);
				try{buffers.close();} catch (SQLException e){}
				return bufferNames;
			} catch (SQLException e) {
				//FIXME close result sets
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	public synchronized FixedArrayNavigableSet<String> bufferNames(final DTSA_BufferType buffer_type, final String module){
		checkBufferType(buffer_type);
		ResultSet buffers;
		while(true){
			try {
				this.sql_translation.sqlSetGetBufferMetaEntriesOfTypeAndModule(this.getBufferSimpleEntriesOfTypeAndModule, buffer_type, module);
				buffers = this.getBufferSimpleEntriesOfTypeAndModule.executeQuery();
				FixedArrayNavigableSet<String> bufferNames = this.sql_translation.sqlReadBufferMetaEntriesNames(buffers);
				try{buffers.close();} catch (SQLException e){}
				return bufferNames;
			} catch (SQLException e) {
				//FIXME close result sets
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	public synchronized FixedArrayNavigableMap<String, String> bufferNameSQLMappings(final DTSA_BufferType buffer_type){
		checkBufferType(buffer_type);
		while(true){
			try {
				this.sql_translation.sqlSetGetBufferMetaEntriesOfType(this.getBufferSimpleEntriesOfType, buffer_type);
				ResultSet buffers = this.getBufferSimpleEntriesOfType.executeQuery();
				//FIXME close result sets
				return this.sql_translation.sqlReadBufferMetaEntriesSQLNameMapping(buffers);
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	public synchronized FixedArrayNavigableMap<String, String> bufferNameSQLMappings(final DTSA_BufferType buffer_type, final String module){
		checkBufferType(buffer_type);
		while(true){
			try {
				this.sql_translation.sqlSetGetBufferMetaEntriesOfTypeAndModule(this.getBufferSimpleEntriesOfTypeAndModule, buffer_type, module);
				ResultSet buffers = this.getBufferSimpleEntriesOfTypeAndModule.executeQuery();
				//FIXME close result sets
				return this.sql_translation.sqlReadBufferMetaEntriesSQLNameMapping(buffers);
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	public synchronized String bufferNameSQLMapping(final DTSA_BufferType buffer_type, final String buffer_name){
		checkBufferType(buffer_type);
		while(true){
			try {
				this.sql_translation.sqlSetGetBufferMetaEntry(this.getBufferSimpleEntry, buffer_type, buffer_name);
				ResultSet buffer = this.getBufferSimpleEntry.executeQuery();
				if(buffer.next()){
					String bufferNameSQLMapping = this.sql_translation.sqlReadBufferMetaEntryTable(buffer);
					try{buffer.close();} catch (SQLException e){}
					return bufferNameSQLMapping;
				}
				else{
					try{buffer.close();} catch (SQLException e){}
					throw new NoSuchElementException("No such buffer: " + buffer_type + " " + buffer_name);
				}
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	public synchronized boolean existsBuffer(final DTSA_BufferType buffer_type, final String buffer_name){
		checkBufferType(buffer_type);
		while(true){
			try {
				this.sql_translation.sqlSetGetBufferMetaEntry(this.getBufferSimpleEntry, buffer_type, buffer_name);
				ResultSet buffer = this.getBufferSimpleEntry.executeQuery();
				if(buffer.next()){
					return true;
				}
				else{
					return false;
				}
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	public synchronized int bufferId(final DTSA_BufferType buffer_type, final String buffer_name){
		checkBufferType(buffer_type);
		while(true){
			try {
				this.sql_translation.sqlSetGetBufferMetaEntry(this.getBufferSimpleEntry, buffer_type, buffer_name);
				ResultSet buffer = this.getBufferSimpleEntry.executeQuery();
				if(buffer.next()){
					return this.sql_translation.sqlReadBufferMetaEntryId(buffer);
				}
				else{
					throw new NoSuchElementException("A buffer of type \"" + buffer_type.toString() + "\" and with name \"" + buffer_name + "\" does not exist.");
				}
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	@Override
	public synchronized FixedArrayNavigableSet<String> attributeNames(final DTSA_BufferType buffer_type, final String buffer_name){
		checkBufferType(buffer_type);
		if(!this.existsBuffer(buffer_type, buffer_name)){
			throw new NoSuchElementException("A buffer of type \"" + buffer_type.toString() + "\" and with name \"" + buffer_name + "\" does not exist.");
		}
		while(true){
			try {
				this.sql_translation.sqlSetGetAttributeMetaEntries(this.getAttributeEntries, buffer_type, buffer_name);
				ResultSet attributes = this.getAttributeEntries.executeQuery();
				return this.sql_translation.sqlReadAttributeMetaEntriesNames(attributes).filter(relevantAttributeFilter);
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	public synchronized FixedArrayNavigableMap<String, String> attributeNameSQLMapping(final DTSA_BufferType buffer_type, final String buffer_name){
		return this.unfilteredAttributeNameSQLMapping(buffer_type, buffer_name).filter(relevantAttributeFilter);
	}
	
	public synchronized FixedArrayNavigableMap<String, String> unfilteredAttributeNameSQLMapping(final DTSA_BufferType buffer_type, final String buffer_name){
		checkBufferType(buffer_type);
		if(!this.existsBuffer(buffer_type, buffer_name)){
			throw new NoSuchElementException("A buffer of type \"" + buffer_type.toString() + "\" and with name \"" + buffer_name + "\" does not exist.");
		}
		while(true){
			try {
				this.sql_translation.sqlSetGetAttributeMetaEntries(this.getAttributeEntries, buffer_type, buffer_name);
				ResultSet attributes = this.getAttributeEntries.executeQuery();
				return this.sql_translation.sqlReadAttributeMetaEntriesSQLNameMapping(attributes);
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	public synchronized String attributeNameSQLMapping(final DTSA_BufferType buffer_type, final String buffer_name, final String attribute_name){
		checkBufferType(buffer_type);
		if(!this.existsBuffer(buffer_type, buffer_name)){
			throw new NoSuchElementException("A buffer of type \"" + buffer_type.toString() + "\" and with name \"" + buffer_name + "\" does not exist.");
		}
		while(true){
			try {
				this.sql_translation.sqlSetGetAttributeMetaEntry(this.getAttributeEntry, buffer_type, buffer_name, attribute_name);
				ResultSet attribute = this.getAttributeEntry.executeQuery();
				if(attribute.next()){
					return this.sql_translation.sqlReadAttributeMetaEntrySQLName(attribute);
				}
				else{
					throw new IllegalArgumentException("No such attribute in buffer: " + buffer_type + ", " + buffer_name + ", " + attribute_name);
				}
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	@Override
	public SinkStream newSinkStream(final String buffer_name) {
		return new SQL_SinkStream(buffer_name, this);
	}
	
	@Override
	public SourceStream newSourceStream(final String buffer_name) {
		return new SQL_SourceStream(buffer_name, this);
	}
	
	@Override
	public synchronized FixedArrayNavigableMap<String, BasicTypeID> attributeTypes(final DTSA_BufferType buffer_type, final String buffer_name){
		checkBufferType(buffer_type);
		if(!this.existsBuffer(buffer_type, buffer_name)){
			throw new NoSuchElementException("A buffer of type \"" + buffer_type.toString() + "\" and with name \"" + buffer_name + "\" does not exist.");
		}
		while(true){
			try {
				this.sql_translation.sqlSetGetAttributeMetaEntries(this.getAttributeEntries, buffer_type, buffer_name);
				ResultSet attributes = this.getAttributeEntries.executeQuery();
				return this.sql_translation.sqlReadAttributeMetaEntriesSQLTypeMapping(attributes).filter(relevantAttributeFilter);
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	@Override
	public synchronized BasicTypeID attributeType(final DTSA_BufferType buffer_type, final String buffer_name, final String attribute_name){
		checkBufferType(buffer_type);
		if(!this.existsBuffer(buffer_type, buffer_name)){
			throw new NoSuchElementException("A buffer of type \"" + buffer_type.toString() + "\" and with name \"" + buffer_name + "\" does not exist.");
		}
		while(true){
			try {
				this.sql_translation.sqlSetGetAttributeMetaEntry(this.getAttributeEntry, buffer_type, buffer_name, attribute_name);
				ResultSet attribute = this.getAttributeEntry.executeQuery();
				if(attribute.next()){
					return this.sql_translation.sqlReadAttributeMetaEntryType(attribute);
				}
				else{
					throw new IllegalArgumentException("No such attribute in buffer: " + buffer_type + ", " + buffer_name + ", " + attribute_name);
				}
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	/*public synchronized Iterable<ResultSet> outputStream(){
		
	}*/
	
	public synchronized Statement createStatement() throws SQLException{
		return this.connection.createStatement();
	}
	
	public synchronized PreparedStatement prepareStatement(final String sql) throws SQLException{
		return this.sql_translation.prepareStatement(this.connection, sql, this.usePreparedStatements);
	}
	
	public synchronized String inputSQL(final String buffer_name, final int bulkSize){
		String input_schema = this.inputSchema();
		String table_name = this.bufferNameSQLMapping(DTSA_BufferType.INPUT, buffer_name);
		FixedArrayNavigableMap<String, String> allAttributes = this.unfilteredAttributeNameSQLMapping(DTSA_BufferType.INPUT, buffer_name);
		String blockAttributeSQLMapping = allAttributes.get(DEFAULTS._o_block);
		int buffer_id = this.bufferId(DTSA_BufferType.INPUT, buffer_name);

		String[] sql_main_attributes = allAttributes.filter(relevantAttributeFilter).values().toArray(new String[0]);
		
		return this.sql_translation.sqlInsertStatement(this.program_meta_schema, input_schema, table_name, buffer_id, bulkSize, blockAttributeSQLMapping, sql_main_attributes);
	}
	
	public synchronized PreparedStatement inputStatement(final String buffer_name, final int bulkSize) throws SQLException{
		return this.prepareStatement(this.inputSQL(buffer_name, bulkSize));
	}
	
	public synchronized String outputSQL(final String buffer_name){
		String output_schema = this.outputSchema();
		String table_name = this.bufferNameSQLMapping(DTSA_BufferType.OUTPUT, buffer_name);
		FixedArrayNavigableMap<String, String> allAttributes = this.attributeNameSQLMapping(DTSA_BufferType.OUTPUT, buffer_name);
		String tsa_range_attribute = DEFAULTS._o_id;//FIXME more generic/hard to maintain
		String sql_range_attribute = this.attributeNameSQLMapping(DTSA_BufferType.OUTPUT, buffer_name, tsa_range_attribute);
		//int range_attribute_index = allAttributes.keys().indexOf(tsa_range_attribute);
		//assert range_attribute_index >= 0;
		//String sql_range_attribute = allAttributes.values().get(range_attribute_index);
		//String[] sql_main_attributes = new String[allAttributes.size()];
		//String[] sql_main_attributes = new String[allAttributes.size() - 1]; //without .id attribute
		String[] sql_main_attributes = allAttributes.values().toArray(new String[allAttributes.size()]);
		//allAttributes.values().copyTo(0, sql_main_attributes, 0, range_attribute_index);
		//allAttributes.values().copyTo(range_attribute_index, sql_main_attributes, range_attribute_index, sql_main_attributes.length - range_attribute_index);
		//allAttributes.values().copyTo(range_attribute_index + 1, sql_main_attributes, range_attribute_index, sql_main_attributes.length - range_attribute_index); //without .id attribute		
		return this.sql_translation.sqlReadRangeStatement(output_schema, table_name, sql_range_attribute, false, true, sql_main_attributes);
	}
	
	public synchronized PreparedStatement outputStatement(final String buffer_name) throws SQLException{
		return this.prepareStatement(this.outputSQL(buffer_name));
	}
	
	public synchronized FixedArrayNavigableMap<String, String> outputBufferQueryOutputMinTimesValueSQLNameMapping(final String buffer_name){
		while(true){
			try {
				this.sql_translation.sqlSetGetOutputQueryMetaEntry(this.getOutputQueryEntry, buffer_name);
				ResultSet buffer = this.getOutputQueryEntry.executeQuery();
				if(buffer.next()){
					FixedArrayNavigableMap<String, String> minTimesValueSQLNameMapping = this.sql_translation.sqlReadQueryMetaEntryOutputStreamBoundsSQLNameMapping(buffer);
					try{buffer.close();} catch(SQLException e){};
					return minTimesValueSQLNameMapping;
				}
				else{
					try{buffer.close();} catch(SQLException e){};
					throw new IllegalArgumentException("No output query for buffer with name \"" + buffer_name + "\" exists.");
				}
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.connect();
					continue;
				}
				else if(SQL_EXCEPTIONS.isRollback(e)){
					this.debug("Statement rolled back. Trying again.", e);
				}
				else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
					this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
					this.initStatements();
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	public synchronized String _o_Id_MaxReadValueSQL(final String buffer_name){
		String _o_id_MinTimeValueSQLName = this.outputBufferQueryOutputMinTimesValueSQLNameMapping(buffer_name).get(TSA_SyntaxPatterns.streamBoundQualifiedName(buffer_name, DEFAULTS._o_id));
		assert _o_id_MinTimeValueSQLName != null;
		
		Persister persister = new Persister();
		de.lmu.ifi.pms.java.util.string.StringBuilder sb = new de.lmu.ifi.pms.java.util.string.StringBuilder();
		
		final String outputs;
		try {
			persister.write(new FixedArrayNavigableSet<String>(buffer_name), sb);
		} catch (Exception e) {
			throw new Error(e);
		}
		outputs = sb.toString();
		
		return "SELECT \"" + _o_id_MinTimeValueSQLName + "\" FROM \"" + this.program_meta_schema + "\".\"queries\" WHERE \"output_type\" = '" + DTSA_BufferType.OUTPUT.name() + "' AND \"outputs\" = '" + outputs + "' ;";
	}
	
	public synchronized PreparedStatement _o_Id_MaxReadValue(final String buffer_name) throws SQLException{
		return this.prepareStatement(this._o_Id_MaxReadValueSQL(buffer_name));
	}
	
	public synchronized String readCurrentBufferContentSQL(final DTSA_BufferType buffer_type, final String buffer_name){
		return this.sql_translation.sqlReadBufferContent(this.inputSchema(), this.bufferNameSQLMapping(buffer_type, buffer_name));
	}
	
	@Override
	public synchronized void close(){
		try {this.statement.close();} catch (SQLException e) {}
		try {this.connection.close();} catch (SQLException e) {}
		this.statement = null;
		this.connection = null;
	}
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(SQL_EngineIO.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
	public static final Adapter<String, String> fromCamelCaseToUndescoreAdapter = new Adapter<String, String>() {
		@Override
		public String adapt(final String source) {
			return fromCamelCaseToUndescore(source);
		}
	};
	
	public static String fromCamelCaseToUndescore(final String s){
		return s.replaceAll("([A-Z])", "_$1").substring(1).toLowerCase();
	}
	
	public String[] copyStaticDataFromExistingTables(final String source_schema, final String module, final Adapter<String, String> bufferToExternalTableNameConversion, final boolean useShortNames){
		FixedArrayNavigableSet<String> buffersInModule = this.bufferNames(DTSA_BufferType.INPUT, module);
		String[] sqlQueries = new String[buffersInModule.size()];
		for(int index = 0; index < sqlQueries.length; ++index){
			String buffer_name = buffersInModule.get(index);
			String external_table_name;
			if(useShortNames){
				external_table_name = TSA_SyntaxPatterns.bufferFullName(buffer_name)[1];
			}
			else{
				external_table_name = buffer_name;
			}
			sqlQueries[index] = this.copyStaticDataFromExistingTable(source_schema, bufferToExternalTableNameConversion.adapt(external_table_name), buffer_name);
		}
		return sqlQueries;
	}
	
	public String copyStaticDataFromExistingTable(final String source_schema, final String source_table, final String buffer_name){
		return this.copyStaticDataFromExistingTable(source_schema, source_table, buffer_name, this.attributeNames(DTSA_BufferType.INPUT, buffer_name));
	}
	
	public String copyStaticDataFromExistingTable(final String source_schema, final String source_table, final String buffer_name, final String... attributes){		
		return this.copyStaticDataFromExistingTable(source_schema, source_table, buffer_name, new FixedArrayNavigableSet<String>(attributes));
	}
	
	public String copyStaticDataFromExistingTable(final String source_schema, final String source_table, final String buffer_name, final FixedArrayNavigableSet<String> attributes){		
		return this.copyStaticDataFromExistingTable(source_schema, source_table, buffer_name, new FixedArrayNavigableMap<String, String>(attributes, attributes));
	}

	public String copyStaticDataFromExistingTable(final String source_schema, final String source_table, final String buffer_name, final FixedArrayNavigableMap<String, String> attributeDestSourceMapping){
		
		String input_schema = this.inputSchema();
		String buffer_name_SQL_mapping = this.bufferNameSQLMapping(DTSA_BufferType.INPUT, buffer_name);
		FixedArrayNavigableMap<String, String> attributeSQLMapping = this.attributeNameSQLMapping(DTSA_BufferType.INPUT, buffer_name);
		
		StringBuilder sb = new StringBuilder();
		sb.append("INSERT INTO \"");
		sb.append(input_schema);
		sb.append("\".\"");
		sb.append(buffer_name_SQL_mapping);
		sb.append("\" (");
		
		assert attributeDestSourceMapping.size() > 0;
		for(String attribute: attributeDestSourceMapping.keys()){
			sb.append("\"");
			sb.append(attributeSQLMapping.get(attribute));
			sb.append("\", ");
		}
		sb.setLength(sb.length() - ", ".length());
			
		sb.append(") SELECT ");
		
		for(String attribute: attributeDestSourceMapping.values()){
			sb.append("\"");
			sb.append(attribute);
			sb.append("\", ");
		}
		sb.setLength(sb.length() - ", ".length());
		
		sb.append(" FROM \"");
		sb.append(source_schema);
		sb.append("\".\"");
		sb.append(source_table);
		sb.append("\" WHERE ");
		
		for(String attribute: attributeDestSourceMapping.values()){
			sb.append("\"");
			sb.append(attribute);
			sb.append("\" IS NOT NULL AND ");
		}
		sb.setLength(sb.length() - " AND ".length());
		
		sb.append(";");
		return sb.toString();
	}
}
