package de.lmu.ifi.pms.cep.tsa.time.analysis;

import de.lmu.ifi.pms.cep.tsa.time.exceptions.MatrixNotQuadradicException;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;

public class SimpleFloydWarshall extends TemporalDistanceCalculationAlgorithm {

	long[][] matrix;
	boolean isConsistent;
	int numberOfAttributes;
	int length;
	
	boolean isComputed;
	
	public SimpleFloydWarshall(final long[][] matrix, final int numberOfAttributes){
		this.matrix = matrix;//TODO copy?
		this.length = this.matrix.length;
		
		this.isConsistent = true;
		
		this.numberOfAttributes = numberOfAttributes;
		
		for(int row = 0; row < this.length; ++row){
			if(this.length!=this.matrix[row].length){
				throw new MatrixNotQuadradicException("Matrix not quadratic!");
			}//TODO better exception (rows with different length)
		}
		this.isComputed = false;
	}
	
	@Override
	public void calculate(){
		for(int k=this.length - 1; k >= this.numberOfAttributes; --k){
			if(this.matrix[k][k] < ArithmeticAnalyzer.ZERO){
				this.isConsistent = false;
				this.isComputed = true;
				return;
			}
			for(int i=0;i<k;++i){
				for(int j=0;j<k;++j){												
					long m = this.matrix[i][k];
					long n = this.matrix[k][j];
					long k1 = m + n;
					long overflow = ((m ^ k1) & (n ^ k1)) >> 63;
					//overflow occured if overflow = -1 (only 1 bits) else overflow = 0 (only 0 bits);		
					long plusInfty = (m << 62 | n << 62 | (overflow & k1)) >> 63;
					//is -1 if n or m are PLUS_INFTY or an overflow occurs else 0
					//if neither m nor n is PLUS_INFTY then the second bit of n and m is 0 and therefore the last two bits of k are either 00 or 01 or 10 
					//in the case of 00 or 01 either a overflow occured or k is smaller than PLUS_INFTY
					//in the case of 10 either a overflow occured or k is smaller or equal than PLUS_INFTY. 
					//If k is equal to PLUS_INFTY then the last two bits (01) mean that both input values were not reduced by epsilon. 
					//Thus the last two bits of the result should be 01, i.e. the result is PLUS_INFTY - 1.		
					long minusInfty = overflow & m & ArithmeticAnalyzer.MINUS_INFTY;//(overflow & m) >> 63 & MINUS_INFTY would be independent from the value of MINUS_INFTY
					//is MINUS_INFTY if an underflow occurs else 0
					long sum = (plusInfty & ArithmeticAnalyzer.PLUS_INFTY) | minusInfty | (~overflow & ~plusInfty & ((k1 & ArithmeticAnalyzer.CONST_MASK) | (k1 & ArithmeticAnalyzer.EPSILON_MASK) >>> 1));
					this.matrix[i][j] = sum < this.matrix[i][j] ? sum : this.matrix[i][j];
				}
			}
		}
		for(int k=this.numberOfAttributes - 1 ; k >= 0; --k){
			if(this.matrix[k][k] < 0){
				this.isConsistent = false;
				this.isComputed = true;
				return;
			}
			for(int i=0;i<this.numberOfAttributes;++i){
				for(int j=0;j<this.numberOfAttributes;++j){						
					long m = this.matrix[i][k];
					long n = this.matrix[k][j];
					long k1 = m + n;
					long overflow = ((m ^ k1) & (n ^ k1)) >> 63;
					//overflow occured if overflow = -1 (only 1 bits) else overflow = 0 (only 0 bits);		
					long plusInfty = (m << 62 | n << 62 | (overflow & k1)) >> 63;
					//is -1 if n or m are PLUS_INFTY or an overflow occurs else 0
					//if neither m nor n is PLUS_INFTY then the second bit of n and m is 0 and therefore the last two bits of k are either 00 or 01 or 10 
					//in the case of 00 or 01 either a overflow occured or k is smaller than PLUS_INFTY
					//in the case of 10 either a overflow occured or k is smaller or equal than PLUS_INFTY. 
					//If k is equal to PLUS_INFTY then the last two bits (01) mean that both input values were not reduced by epsilon. 
					//Thus the last two bits of the result should be 01, i.e. the result is PLUS_INFTY - 1.		
					long minusInfty = overflow & m & ArithmeticAnalyzer.MINUS_INFTY;//(overflow & m) >> 63 & MINUS_INFTY would be independent from the value of MINUS_INFTY
					//is MINUS_INFTY if an underflow occurs else 0
					long sum = (plusInfty & ArithmeticAnalyzer.PLUS_INFTY) | minusInfty | (~overflow & ~plusInfty & ((k1 & ArithmeticAnalyzer.CONST_MASK) | (k1 & ArithmeticAnalyzer.EPSILON_MASK) >>> 1));
					this.matrix[i][j] = sum < this.matrix[i][j] ? sum : this.matrix[i][j];
				}
			}
		}
		this.isComputed = true;
	}
	
	@Override
	public long[][] getMatrix(){
		if(this.isComputed){
			return this.matrix;
		}
		else{
			throw new IllegalStateException();
		}
	}
	
	@Override
	public boolean isConsistent(){
		if(this.isComputed){
			return this.isConsistent;
		}
		else{
			throw new IllegalStateException();
		}
	}
	
	@Override
	public String toString(){
		StringBuilder sb = new StringBuilder();
		String[] names = new String[this.matrix.length];
		for(int index = 0; index < names.length; ++index){
			names[index] = "\"" + index + "\"";
		}
		
		SimpleArithmeticAnalyzer.writeMatrix(sb, new FixedArray<String>(false, names), this.matrix);
		return sb.toString();
	}
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	

}
