package de.lmu.ifi.pms.java.util.stream.io;

import java.io.Closeable;
import java.util.NoSuchElementException;

import de.lmu.ifi.pms.java.lang.iteration.AbstractFixedIterator;
import de.lmu.ifi.pms.java.lang.iteration.AbstractIterable;
import de.lmu.ifi.pms.java.util.Buffer;
import de.lmu.ifi.pms.java.util.stream.Source;
import de.lmu.ifi.pms.java.util.stream.SourceStream;
import de.lmu.ifi.pms.java.util.tupel.Single;


public class BlockingPipe<Type> extends AbstractIterable<Type> implements SourceStream<Type>{
	
	private boolean open = true;

	protected final Buffer<Type> buffer;
	
	protected final Object lock;
	
	public BlockingPipe(){
		this(256);
	}
	
	public BlockingPipe(final int bufferCapacity){
		this.buffer = new Buffer<Type>(bufferCapacity);
		this.lock = this.buffer;
	}
	
	public BlockingPipe(final int bufferCapacity, final Object lock){
		this.buffer = new Buffer<Type>(bufferCapacity);
		this.lock = lock;
	}
	
	protected void waitWhileEmpty(){
		synchronized(this.lock){
			while(this.buffer.isEmpty() && this.isOpen()){
				try {
					this.lock.wait();
				} catch (final InterruptedException e) {

				}
			}
		}
	}
	
	protected void waitWhileFull(){
		synchronized(this.lock){
			while(this.buffer.isFull() && this.isOpen()){
				try {
					this.lock.wait();
				} catch (final InterruptedException e) {

				}
			}
		}
	}
	
	public boolean hasNext(){
		if(this.isOpen()){
			synchronized(this.lock){
				this.waitWhileEmpty();
				return this.isOpen();
			}	
		}
		else{
			throw new IllegalStateException("Already closed.");
		}
	}
	
	public Type readNext(){
		if(this.isOpen()){
			synchronized(this.lock){
				this.waitWhileEmpty();
				if(this.isOpen()){
					final Type e = this.buffer.first();
					this.buffer.removeFirst();
					this.lock.notifyAll();
					return e;
				}
				else{
					this.lock.notifyAll();
					throw new NoSuchElementException();
				}
			}
		}
		else{
			throw new IllegalStateException("Already closed.");
		}
	}
	
	public Type examineNext(){
		if(this.isOpen()){
			synchronized(this.lock){
				this.waitWhileEmpty();
				if(this.isOpen()){
					return this.buffer.first();
				}
				else{
					this.lock.notifyAll();
					throw new NoSuchElementException();
				}
			}
		}
		else{
			throw new IllegalStateException("Already closed.");
		}
	}
	
	public void writeNext(final Type e){
		if(this.isOpen()){
			synchronized(this.lock){
				this.waitWhileFull();
				if(this.isOpen()){
					this.buffer.addLast(e);
					this.lock.notifyAll();
				}
				else{
					throw new IllegalStateException("Already closed.");
				}
			}
		}
		else{
			throw new IllegalStateException("Already closed.");
		}
	}
	
	@Override
	public void close(){
		synchronized(this.lock){
			synchronized(this){
				this.open = false;
			}
			this.lock.notifyAll();
		}
	}
	
	public void closeWhenEmpty(){
		final Thread thread = new Thread(){
			@Override
			public void run() {
				synchronized(BlockingPipe.this.lock){
					while(!BlockingPipe.this.buffer.isEmpty() && BlockingPipe.this.isOpen()){
						try {
							BlockingPipe.this.lock.wait();
						} catch (final InterruptedException e) {

						}
					}
					BlockingPipe.this.close();
				}
			}
		};
		thread.setDaemon(true);
		thread.start();

	}
	
	public boolean isOpen(){
		synchronized (this) {
			return this.open;
		}
	}
	
	@Override
	public Source<Type> iterator() {
		final BlockingPipe<Type> that = this;
		return new AbstractFixedIterator<Type>(){
			@Override
			public boolean hasNext() {
				try{
					return that.hasNext();
				}
				catch(final IllegalStateException e){
					return false;
				}
			}
			
			@Override
			public Type next() {
				return BlockingPipe.this.readNext();
			}

		};
	}
	
	@Override
	public Source<Type> source() {
		return this.iterator();
	}
	
	public synchronized final int capacity(){
		return this.buffer.size() + this.buffer.remainingCapacity();
	}
	
	public ObjectInputStream<Type, ? extends Single<Closeable,BlockingPipe<Type>>> getInputStream(){
		return new LocalObjectInputStream<Type>(this);
	}
	
	public ObjectOutputStream<Type, ? extends Single<Closeable,BlockingPipe<Type>>> getOutputStream(){
		return new LocalObjectOutputStream<Type>(this);
	}
	
}
